import os as _os
import time as _time
from typing import (
    Any as _Any,
    List as _List,
)

from mykit.kit.fileops.simple import (
    same_ext_for_all_dir_files as _same_ext_for_all_dir_files,
    list_dir as _list_dir
)
from mykit.kit.path import SafeJSON as _SafeJSON
from mykit.kit.time import TimeFmt as _TimeFmt


class JsonListDbManager:
    """
    A simple database system that uses JSON file to store data as a list

    ---

    ## Demo
    
    The database folder:
    ```
    ## Note: the actual name would be `db_YYYYMMDD_HHMMSS.json`; below is just for conciseness.
    container/
    ├── db_1.json
    └── db_2.json
    ```
    
    The database file (`container/db_2.json`):
    ```
    ['x1', 'x2', 'and so on..']
    ```
    
    Let's say the latest database file (`db_2.json`) contains 3 pieces of
    data (3 items inside the list). Now, we want to set the maximum data
    per database file to 3. If the next data is intended to be added
    to `db_2.json`, it won't be added because the maximum capacity has
    been reached. Instead, a new database file (`db_3.json`) will be
    created to store that new data.
    """
    
    def __init__(self, container:str, /) -> None:
        """
        ### Params
        - `container`: the absolute path to the folder containing the database files

        ### Exceptions
        - `NotADirectoryError`: if `container` is not a directory
        - `AssertionError`: if an item in `container` is not a JSON file
        - Please manually review this function dependencies for more exceptions
        """
        if not _os.path.isdir(container): raise NotADirectoryError(f'Not a dir: {repr(container)}.')
        try:
            if not _same_ext_for_all_dir_files(container, '.json'):
                raise AssertionError(f'All items in {repr(container)} must be JSON files.')
        except (NotADirectoryError, ValueError, AssertionError) as err:
            raise AssertionError(err)
        self.container = container
    
    def _get_blocks(self):
        blocks = _list_dir(self.container)  # Each DB file is called "block"
        blocks = sorted(blocks)  # Sorted from oldest to newest: [oldest, ..., newest]
        return blocks

    @property
    def num_blocks(self) -> int:
        """Number of database files"""
        return len(self._get_blocks())

    def save(self, data:_Any, /, max:int=1024) -> None:
        """
        ### Params:
        - `data`: the new data to be added
        - `max`: maximum data per database file (items inside the list).
                 Will be clamped to at least `0`. `0` and `1` are valid,
                 even though they are truly rarely used.
        """
        blocks = self._get_blocks()

        if blocks == []:  # Handle init case: empty database
            target = _os.path.join(self.container, f'db_{_TimeFmt.sort()}.json')
            db = [data]
            _SafeJSON.write(target, db, do_log=False)
        else:
            target = blocks[-1][1]  # Absolute path to the newest database file
            db = _SafeJSON.read(target)
            if len(db) >= max:
                target = _os.path.join(self.container, f'db_{_TimeFmt.sort()}.json')
                db = [data]
                _SafeJSON.write(target, db, do_log=False)
            else:
                db.append(data)
                _SafeJSON.rewrite(target, db, do_log=False)

    def bulk_save(self, list_of_data:_List[_Any], /, max:int=1024) -> None:
        """Similar to `.save`, but for bulk operations."""
        ## TODO: Implement a much more efficient approach soon
        for data in list_of_data:
            self.save(data, max=max)
            _time.sleep(1/max)  # To prevent collisions in database file names (which happen during testing when `max` is small)

    def get_partial(self, idx:int) -> _List[_Any]:
        """
        Get just a specific database file.

        @param `idx`: `0` for the first file, `1` second, and so on

        ### Exceptions
        - `ValueError`: if `idx` is out of range
        """
        if (idx < 0) or (idx > (self.num_blocks-1)): raise ValueError('Value `idx` out of range.')
        blocks = self._get_blocks()
        pth = blocks[idx][1]
        return _SafeJSON.read(pth)

    def get_all(self) -> _List[_Any]:
        """Get the entire database"""
        out = []
        for _, pth in self._get_blocks():
            out += _SafeJSON.read(pth)
        return out


import math as _math
import random as _random
import time as _time
from typing import (
    List as _List,
    Tuple as _Tuple
)


def _non_periodic_fn(__x: float, /, a: float = 1, b: float = 1, c: float = 1, d: float = 1, e: float = 1, f: float = 1, g: float = 1):
    """
    proven non-periodic.
    ref: https://stackoverflow.com/questions/8798771/perlin-noise-for-1d

    ---

    `__x` is in radian
    """
    return a*(b*_math.sin(c*__x) + d*_math.sin(e*_math.pi*__x) + f*_math.sin(g*_math.e*__x))

def random_sine_noise(
    seed: float,
    nsample: int,
    ymin: int,
    ymax: int,
    params: _Tuple = (1, 1, 1, 1, 1, 1, 1)
) -> _List[_Tuple[float, float]]:

    _random.seed(seed)
    shift = _random.randint(-1000000000, 1000000000)

    points = []
    for x in range(nsample):
        x += shift
        points.append((x, _non_periodic_fn(x*_math.pi/180, *params)))

    ## adjusting the y interval
    y_values = [point[1] for point in points]
    ymin_current = min(y_values)
    ymax_current = max(y_values)
    scale_factor = (ymax - ymin) / (ymax_current - ymin_current)
    points = [(x, ymin + (y - ymin_current)*scale_factor) for (x, y) in points]

    return points


def _pn1d_noise(x: int, seed: int) -> float:
    """Generate deterministic noise based on the given seed."""
    x = (x << 13) ^ x ^ seed
    return 1.0 - ((x * (x * x * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0

def _pn1d_noise_smoothed(x: int, seed: int) -> float:
    """combines multiple noise samples to compute the final value"""
    return _pn1d_noise(x, seed)/2 + _pn1d_noise(x - 1, seed)/4 + _pn1d_noise(x + 1, seed)/4

def _pn1d_noise_interpolated(x: int, seed: int) -> float:
    """using cubic interpolation"""
    int_x = int(x)
    frac_x = x - int_x

    v0 = _pn1d_noise_smoothed(int_x - 1, seed)
    v1 = _pn1d_noise_smoothed(int_x, seed)
    v2 = _pn1d_noise_smoothed(int_x + 1, seed)
    v3 = _pn1d_noise_smoothed(int_x + 2, seed)

    ## cubic interpolation
    p = (v3 - v2) - (v0 - v1)
    q = (v0 - v1) - p
    r = v2 - v0
    s = v1
    return p*frac_x*frac_x*frac_x + q*frac_x*frac_x + r*frac_x + s

def perlin_noise_1d(x: float, /, persistence: float = 0.5, octaves: int = 1, frequency: int = 2, seed: int = _time.time()) -> float:
    """
    Generate 1D Perlin noise for the given input parameters.
    ref: https://web.archive.org/web/20160530124230/http://freespace.virgin.net/hugo.elias/models/m_perlin.htm

    ---

    ## params
        `x`: input value.
        `persistence`: persistence value determining the amplitude decrease per octave.
        `octaves`: number of octaves or levels of detail in the Perlin noise.
        `frequency`: lower frequency results in higher detail.
        `seed`: seed value for randomness.

    ## returns
        float: Perlin noise value
    """
    total = 0
    for i in range(octaves):
        f = frequency**i
        amplitude = persistence**i
        total += amplitude*_pn1d_noise_interpolated(x*f, seed)
    return total



## dev-docs: This is the variation and the next generation of mykit.kit.path.SafeJSON.
##           SafeJSON will be deprecated soon. TODO: Make NiceJSON, which is the next generation of it.

import json as _json
import os as _os
from typing import (
    Any as _Any,
)

from mykit.kit.fileops.simple import (
    norm_pth as _norm_pth,
    dont_worry_the_path_ends_with as _dont_worry_the_path_ends_with,
    definitely_a_file as _definitely_a_file,
    definitely_a_dir as _definitely_a_dir,
    this_path_must_not_exist as _this_path_must_not_exist
)


class NiceJSON:
    """Procedurally strict for performing file operations like read, write, and rewrite"""

    @staticmethod
    def read(file_path:str, /) -> _Any:
        """
        Get the content of the given JSON file path

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        ## Checks
        _definitely_a_file(path_normalized)
        _dont_worry_the_path_ends_with(path_normalized, '.json')

        ## Read
        with open(path_normalized, 'r') as fp:
            out = _json.load(fp)
        
        return out

    @staticmethod
    def write(file_path:str, content:_Any) -> None:
        """
        Write the given `content` into a new JSON file

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        ## Checks
        _dont_worry_the_path_ends_with(path_normalized, '.json')
        _definitely_a_dir(_os.path.dirname(path_normalized))
        _this_path_must_not_exist(path_normalized)

        with open(path_normalized, 'w') as fp:
            _json.dump(content, fp)

    @staticmethod
    def rewrite(file_path:str, content:_Any) -> None:
        """
        Replace the existing JSON file at `file_path` with the new `content`

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        tmp_path = path_normalized + '.tmp'
        bak_path = path_normalized + '.bak'

        ## Checks
        _dont_worry_the_path_ends_with(path_normalized, '.json')
        _definitely_a_file(path_normalized)
        _this_path_must_not_exist(tmp_path)
        _this_path_must_not_exist(bak_path)

        ## Run
        with open(tmp_path, 'w') as fp: _json.dump(content, fp)  # writing the new as temp
        _os.rename(path_normalized, bak_path)                    # backup the previous
        _os.rename(tmp_path, path_normalized)                    # rename the temp to the new
        _os.remove(bak_path)                                     # delete the previous

    @staticmethod
    def recover(file_path:str, /) -> None:
        """
        Try to fix the broken JSON file, mostly because of interrupted or sudden
        stops when performing `NiceJSON.rewrite` above. Note that this is still
        not fully optimized and might still change in the future.

        ## Exceptions
        - Please manually inspect this function for all possible exceptions
        """

        ## Normalize
        path_normalized = _norm_pth(file_path)

        tmp_path = path_normalized + '.tmp'
        bak_path = path_normalized + '.bak'

        ## Checks
        _dont_worry_the_path_ends_with(path_normalized, '.json')


        ## case I
        if _os.path.exists(path_normalized) and _os.path.exists(tmp_path) and (not _os.path.exists(bak_path)):
            _os.remove(tmp_path)
            return

        ## case II
        if (not _os.path.exists(path_normalized)) and _os.path.exists(tmp_path) and _os.path.exists(bak_path):
            _os.rename(tmp_path, path_normalized)
            _os.remove(bak_path)
            return

        ## case III
        if _os.path.exists(path_normalized) and (not _os.path.exists(tmp_path)) and _os.path.exists(bak_path):
            _os.remove(bak_path)
            return

        ## case IV (weirdly happening, but currently disabled because i dont know if this is also happening on different environments)
        # if (not _os.path.exists(path_normalized)) and (not _os.path.exists(tmp_path)) and _os.path.exists(bak_path):
        #     _os.rename(bak_path, path_normalized)
        #     return



import TurboTest as tt
from . import NiceIO

import os
import tempfile


## dev-docs: These are the wrong ways to do the tests since the test data
#            can't be shared because the temp file will exist after just one test.
## <delete this soon>

# ## Temporary dir
# TD = tempfile.mkdtemp()


# def read_successfully_empty_file():
    
#     pth = os.path.join(TD, 'test_file.txt')
#     tt.this_is_False(os.path.isfile(pth))
#     open(pth, 'w').close()
#     tt.this_is_True(os.path.isfile(pth))

#     x = NiceIO.read(pth)
#     tt.both_are_equal(x, '')

# def read_successfully_a_file():
    
#     pth = os.path.join(TD, 'test_file.txt')
#     tt.this_is_False(os.path.isfile(pth))
#     with open(pth, 'w') as f: f.write('abc')
#     tt.this_is_True(os.path.isfile(pth))

#     x = NiceIO.read(pth)
#     tt.both_are_equal(x, 'abc')

# def fail_when_reading_nonexistent_file():
#     pth = os.path.join(TD, 'test_file.txt')
#     with tt.these_will_raise(FileNotFoundError) as its: NiceIO.read()

## </delete this soon>


def read_successfully_empty_file():

    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'test_file.txt')
    tt.this_is_False(os.path.isfile(pth))
    open(pth, 'w').close()
    tt.this_is_True(os.path.isfile(pth))

    x = NiceIO.read(pth)
    tt.both_are_equal(x, '')

def read_successfully_a_file():
    
    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'test_file.txt')
    tt.this_is_False(os.path.isfile(pth))
    with open(pth, 'w') as f: f.write('abc')
    tt.this_is_True(os.path.isfile(pth))

    x = NiceIO.read(pth)
    tt.both_are_equal(x, 'abc')

def fail_when_reading_nonexistent_file():
    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'test_file.txt')
    with tt.these_will_raise(FileNotFoundError) as its: NiceIO.read(pth)
    tt.both_are_equal(its.exception_msg, f"Not a file: {repr(pth)}.")

def fail_if_reading_from_an_unexpected_file_type():
    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'test_file.txt')
    suffixes = '.py'
    open(pth, 'w').close()  # create the file
    with tt.these_will_raise(AssertionError) as its: NiceIO.read(pth, suffixes)
    tt.both_are_equal(its.exception_msg, f"Invalid suffixes: [expected: {repr(suffixes)}] [got: {repr(pth)}]")



def write_successfully_empty_file():

    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'foo.md')
    tt.this_is_False(os.path.isfile(pth))
    NiceIO.write(pth, '')
    tt.this_is_True(os.path.isfile(pth))

    x = NiceIO.read(pth)
    tt.both_are_equal(x, '')

def write_successfully_a_file():

    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'foo.md')
    tt.this_is_False(os.path.isfile(pth))
    NiceIO.write(pth, 'hi\nmom')
    tt.this_is_True(os.path.isfile(pth))

    x = NiceIO.read(pth)
    tt.both_are_equal(x, 'hi\nmom')

def fail_to_write_if_unexpected_file_type():
    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'foo.md')
    suffixes = '.txt'
    with tt.these_will_raise(AssertionError) as its: NiceIO.write(pth, 'abc', suffixes)
    tt.both_are_equal(its.exception_msg, f"Invalid suffixes: [expected: {repr(suffixes)}] [got: {repr(pth)}]")

def fail_to_write_when_the_parent_dir_does_not_exist():
    temp_dir = tempfile.mkdtemp()
    parent_dir = os.path.join(temp_dir, 'nonexistent')
    pth = os.path.join(parent_dir, 'foo.md')
    with tt.these_will_raise(NotADirectoryError) as its: NiceIO.write(pth, 'abc')
    tt.both_are_equal(its.exception_msg, f"Not a dir: {repr(parent_dir)}.")

def fail_to_write_when_the_target_path_already_exists():
    
    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'foo.md')
    
    tt.this_is_False(os.path.isfile(pth))
    open(pth, 'w').close()
    tt.this_is_True(os.path.isfile(pth))

    with tt.these_will_raise(AssertionError) as its: NiceIO.write(pth, '')
    tt.both_are_equal(its.exception_msg, f"Already exists: {repr(pth)}.")



def rewrite_successfully():

    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'a.txt')

    tt.this_is_False(os.path.isfile(pth))
    open(pth, 'w').close()
    tt.this_is_True(os.path.isfile(pth))
    tt.both_are_equal(NiceIO.read(pth), '')

    NiceIO.rewrite(pth, 'foo bar\nbaz')
    tt.both_are_equal(NiceIO.read(pth), 'foo bar\nbaz')

def rewrite_successfully_II():

    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'a.txt')

    tt.this_is_False(os.path.isfile(pth))
    with open(pth, 'w') as f: f.write('before')
    tt.this_is_True(os.path.isfile(pth))
    tt.both_are_equal(NiceIO.read(pth), 'before')

    NiceIO.rewrite(pth, 'after')
    tt.both_are_equal(NiceIO.read(pth), 'after')

def cant_rewrite_if_unexpected_file_type():
    
    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'a.txt')
    suffixes = '.c'

    tt.this_is_False(os.path.isfile(pth))
    with open(pth, 'w') as f: f.write('before')
    tt.this_is_True(os.path.isfile(pth))
    tt.both_are_equal(NiceIO.read(pth), 'before')

    with tt.these_will_raise(AssertionError) as its: NiceIO.rewrite(pth, 'after', suffixes)
    tt.both_are_equal(its.exception_msg, f"Invalid suffixes: [expected: {repr(suffixes)}] [got: {repr(pth)}]")

def cant_rewrite_if_the_file_doesnt_exist():

    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'a.txt')

    tt.this_is_True(not os.path.isfile(pth))

    with tt.these_will_raise(FileNotFoundError) as its: NiceIO.rewrite(pth, '')
    tt.both_are_equal(its.exception_msg, f"Not a file: {repr(pth)}.")

def cant_rewrite_when_previous_rewrite_failed():

    temp_dir = tempfile.mkdtemp()

    ## If failed, so .tmp exists
    ## vvvvvvvvvvvvvvvvvvvvvvvvv

    pth = os.path.join(temp_dir, 'a.txt')
    pth_tmp = pth + '.tmp'
    open(pth, 'w').close()
    open(pth_tmp, 'w').close()
    tt.this_is_True(os.path.isfile(pth))
    tt.this_is_True(os.path.isfile(pth_tmp))
    with tt.these_will_raise(AssertionError) as its: NiceIO.rewrite(pth, '')
    tt.both_are_equal(its.exception_msg, f"Already exists: {repr(pth_tmp)}.")

    ## If failed, so .bak exists
    ## vvvvvvvvvvvvvvvvvvvvvvvvv

    pth = os.path.join(temp_dir, 'b.txt')
    pth_bak = pth + '.bak'
    open(pth, 'w').close()
    open(pth_bak, 'w').close()
    tt.this_is_True(os.path.isfile(pth))
    tt.this_is_True(os.path.isfile(pth_bak))
    with tt.these_will_raise(AssertionError) as its: NiceIO.rewrite(pth, '')
    tt.both_are_equal(its.exception_msg, f"Already exists: {repr(pth_bak)}.")



## TODO: write the test suite for `recover` after it's fully developed
# def recover_



def erase_an_existing_file():

    temp_dir = tempfile.mkdtemp()
    pth = os.path.join(temp_dir, 'my.txt')
    with open(pth, 'w') as f: f.write('hi mom\n123')
    tt.both_are_equal(NiceIO.read(pth), 'hi mom\n123')

    NiceIO.erase(pth)
    tt.both_are_equal(NiceIO.read(pth), '')

## These are good to do but not necessary since the above cases already cover these
# def cant_erase_if_unexpected_file_type(): ...
# def cant_erase_if_the_file_doesnt_exist(): ...
# def cant_erase_when_previous_rewrite_operation_failed(): ...



import time as _time
import tkinter as _tk
from typing import (
    Any as _Any,
    Callable as _Callable,
    Dict as _Dict,
    List as _List,
)

from mykit.app.button import Button as _Button
from mykit.app.label import Label as _Label
from mykit.app.slider import _Slider

from mykit.app.complex.plot import Plot as _Plot
from mykit.app.complex.biplot import Biplot as _Biplot
from mykit.app.arrow import Arrow as _Arrow

from mykit.app.photoimg import PhotoImg


def _install_components(app):
    PhotoImg._install(app)


class App:
    """
    A single-page app framework.

    ## Limitations
    - Currently available only in fullscreen mode
    """

    def __init__(
        self,
        name: str = 'app',
        bg: str = '#111111',
    ) -> None:

        self.root = _tk.Tk()
        self.root.attributes('-fullscreen', True)
        self.root.title(name)

        ## App's main page
        self.page = _tk.Canvas(
            master=self.root,
            width=self.root.winfo_screenwidth(),
            height=self.root.winfo_screenheight(),
            background=bg,
            borderwidth=0, highlightthickness=0
        )
        self.page.place(x=0, y=0)


        ## <register the page>
        
        ## widgets
        _Button._set_page(self.page)
        _Slider._set_page(self.page)

        ## others
        _Arrow._set_page(self.page)
        _Plot._set_page(self.page)
        _Biplot._set_page(self.page)

        ## </register the page>


        _install_components(self)


        ## <constants>

        self.T = _time.time()  # The timestamp when the app instance was initiated, usually used as the startup time.
        self.MON_WIDTH = self.root.winfo_screenwidth()
        self.MON_HEIGHT = self.root.winfo_screenheight()
        
        ## </constants>


        ## <runtime>

        self._left_mouse_press = []
        self._left_mouse_hold = []
        self._left_mouse_release = []

        self._background_processes = {}

        self._after_initialization = []
        self._dependencies = {}

        self._setup = []
        self._teardown = []

        ## </runtime>

    def listen(self, to:str, do:_Callable[[_tk.Event], None]) -> None:
        """
        Add event listener.

        ## Params
        - `to`: event type:
            - `"left-mouse-press"` or `"lmp"`
            - `"left-mouse-hold"` or `"lmh"`
            - `"left-mouse-release"` or `"lmr"`

        ## Docs
        - `do` function takes 1 positional parameter, which is a tkinter event object
        """
        
        if to in {'left-mouse-press', 'lmp'}:
            self._left_mouse_press.append(do)
        elif to in {'left-mouse-hold', 'lmh'}:
            self._left_mouse_hold.append(do)
        elif to in {'left-mouse-release', 'lmr'}:
            self._left_mouse_release.append(do)
        else:
            ValueError(f'Invalid event: {repr(to)}.')

    def add_background_processes(self, every: int, do: _Callable[[], None]) -> None:
        """
        Execute `do` every `every` milliseconds.
        
        ## Docs
        - The first execution occurs immediately after the app runs.
        """
        if every not in self._background_processes:
            self._background_processes[every] = []
        self._background_processes[every].append(do)

    def setup(self, funcs: _List[_Callable[[], None]]) -> None:
        """Running simple-functions right before startup."""
        self._setup = funcs

    def teardown(self, funcs: _List[_Callable[[], None]]) -> None:
        """Running simple-functions right after the app stops."""
        self._teardown = funcs

    def use(self, component: _Callable[..., None], /, dependencies: _Dict[str, _Any]) -> None:
        """This function helps make the dataflow easier to read."""
        component(**dependencies)

    def after_initialization_use(self, component: _Callable[..., None], /, *, with_dependencies: _List[str]) -> None:
        """Use `component` once all dependencies are ready."""
        self._after_initialization.append([component, with_dependencies])

    def add_dependencies(self, dependency: _Any, /) -> None:
        """Adds a dependency and should be used with the `after_initialization_use` method."""
        name = dependency.__name__
        if name in self._dependencies:
            raise ValueError(f'Dependency {repr(name)} is already added.')
        self._dependencies[name] = dependency

    def run(self):

        ## <internal>

        self.listen(to='left-mouse-press', do=_Button._press_listener)
        self.listen(to='left-mouse-press', do=_Slider._press_listener)

        self.listen(to='left-mouse-hold', do=_Slider._hold_listener)

        self.listen(to='left-mouse-release', do=_Button._release_listener)
        self.listen(to='left-mouse-release', do=_Slider._release_listener)

        self.add_background_processes(every=50, do=_Button._hover_listener)
        self.add_background_processes(every=50, do=_Slider._hover_listener)

        ## </internal>


        ## <listeners>

        self.root.bind('<ButtonPress-1>',   lambda e: [f(e) for f in self._left_mouse_press])
        self.root.bind('<B1-Motion>',       lambda e: [f(e) for f in self._left_mouse_hold])
        self.root.bind('<ButtonRelease-1>', lambda e: [f(e) for f in self._left_mouse_release])

        self.root.bind('<Escape>', lambda e: self.root.destroy())

        ## </listeners>


        def run_background_processes():
            def wrapper(dur, funcs):
                def inner():
                    for fn in funcs: fn()
                    self.root.after(dur, inner)
                return inner
            for dur, funcs in self._background_processes.items():
                fn = wrapper(dur, funcs)
                fn()  # start immediately
        run_background_processes()


        ## Initialize components that need to be initialized after all dependencies are ready
        for component, dependencies in self._after_initialization:
            component(self, **{d: self._dependencies[d] for d in dependencies})


        ## Setup
        for function in self._setup:
            function()

        ## Startup
        self.root.mainloop()

        ## Teardown
        for function in self._teardown:
            function()



import unittest

from mykit.kit.time import TimeFmt


class Test__TimeFmt(unittest.TestCase):

    def setUp(self):
        self.timestamp = 1692630123.0
        self.utc_offset = 0

    def test_date(self):
        result = TimeFmt.date(self.timestamp, self.utc_offset)
        expected = 'Aug 21, 2023'
        self.assertEqual(result, expected)

    def test_hour(self):
        result = TimeFmt.hour(self.timestamp, self.utc_offset)
        expected = '15:02:03'
        self.assertEqual(result, expected)

    def test_sort(self):
        result = TimeFmt.sort(self.timestamp, self.utc_offset)
        expected = '20230821_150203'
        self.assertEqual(result, expected)

    def test_neat(self):
        result = TimeFmt.neat(self.timestamp, self.utc_offset)
        expected = '2023 Aug 21, 15:02:03 UTC+0000'
        self.assertEqual(result, expected)

        result = TimeFmt.neat(self.timestamp, 1.75)
        expected = '2023 Aug 21, 16:47:03 UTC+0145'
        self.assertEqual(result, expected)

        result = TimeFmt.neat(self.timestamp, -11.15)
        expected = '2023 Aug 21, 03:53:03 UTC-1109'
        self.assertEqual(result, expected)


if __name__ == '__main__':
    unittest.main()




## Birthdate: Oct 6, 2023
def rcut(input_str:str, unwanted:str, check:bool=False) -> str:
    """
    right-cut (rcut). Trim the ending characters `unwanted` from the given `input_str`.

    ## Params
    - `input_str`: the input string
    - `unwanted`: the characters to be trimmed from the ending (see examples below to learn more)
    - `check`: if `True`, raise an exception (see below) if `input_str` does not end with `unwanted`

    ## Exceptions
    - `ValueError` if `unwanted` is an empty string
    - `AssertionError` if `input_str` doesn't end with `unwanted` when `check` is `True`

    ## Examples
    - `rcut('foo12', '12')` -> `'foo'`
    - `rcut('hi mom 123', ' 123')` -> `'hi mom'`
    - `rcut('hi mom', 'cool')` -> `'hi mom'`

    ## Docs
    - the reason using `rcut` and not `rstrip` is because `'hi 12123'.rstrip('123')` will be `'hi '`
    """

    ## Check
    if unwanted == '': raise ValueError("rcut: `unwanted` shouldn't be an empty string.")

    if input_str.endswith(unwanted):
        return input_str[:-len(unwanted)]
    else:
        if check:
            raise AssertionError("rcut: the given `input_str` does not end with `unwanted`.")
        return input_str



from mykit.kit.color import colored_len as _colored_len
from mykit.kit.readable.paragraph_width import paragraph_width as _paragraph_width


def box(text:str, /) -> str:
    """
    Surround the single/multi-line text `text` with a box.
    This also works for colored text produced by the mykit `Colored` function.

    ## Demo
    >>> box('test 123')
    >>> # ============
    >>> # ||test 123||
    >>> # ============
    """
    w = _paragraph_width(text)
    text_modified = ''
    for line in text.split('\n'):
        text_modified += '||' + line + ' '*(w-_colored_len(line)) + '||\n'
    return '='*(w+4) + '\n' + text_modified + '='*(w+4)



import os
from pathlib import Path as _Path


def _compare(dir, ref, cmp):
    """
    - `dir`: The current iterated directory
    - `ref`: Reference
    - `cmp`: Comparison
    """

    dir_rel = os.path.relpath(dir, ref)
    dir_cmp_pth = os.path.join(cmp, dir_rel)
    if not os.path.isdir(dir_cmp_pth): raise AssertionError('The current iterated directory is missing in the comparison.')

    for thing in os.listdir(dir):
        thing_pth = os.path.join(dir, thing)

        if os.path.isfile(thing_pth):
            file_rel = os.path.relpath(thing_pth, ref)
            file_cmp_pth = os.path.join(cmp, file_rel)
            if not os.path.isfile(file_cmp_pth): raise AssertionError('The current iterated file is missing in the comparison.')
            with open(thing_pth, 'r') as f: text1 = f.read()
            with open(file_cmp_pth, 'r') as f: text2 = f.read()
            if text1 != text2: raise AssertionError('File content differs between ref and cmp directories.')
        else:
            _compare(thing_pth, ref, cmp)


def equaldirs(dir1:_Path, dir2:_Path, /) -> bool:
    """
    Return `True` if `dir1` and `dir2` are equal (recursively; all the
    file and subdir names in both dirs are the same; all files have
    the same content; but note that file metadata (such as modified/created
    date) may not be considered). Return `False` otherwise.

    ---

    ## Params
    - `dir1`: Absolute path to the directory-1
    - `dir2`: Absolute path to the directory-2

    ## Exceptions
    - `NotADirectoryError`: If `dir1` or `dir2` is not a directory

    ## Docs
    - may not work for directories containing non-text items such as images/videos
    """
    
    ## Checks
    if not os.path.isdir(dir1): raise NotADirectoryError(f'Value `dir1` is not a directory: {repr(dir1)}.')
    if not os.path.isdir(dir2): raise NotADirectoryError(f'Value `dir2` is not a directory: {repr(dir2)}.')

    try:
        _compare(dir1, dir1, dir2)  # dir1 side
        _compare(dir2, dir2, dir1)  # dir2 side
    except AssertionError:
        return False
    
    return True


"""
Simple concise program plugins (SCPP).

- This module focuses on UI/UX, aiming for readability and simplicity.
- This module is used along with other mykit-program related tools, like the ones from `cli_parser`, etc.

[20230919] Note for my future self: You will and should regroup the entire `/kit/` master module to make all of these more sense.
"""



import unittest

from mykit.kit.readable.box import box

from mykit.kit.color import Colored, Hex


class Test__box(unittest.TestCase):

    def test_single(self):
        
        text = 'hi 123'
        result = box(text)
        expected = (
            '==========\n'
            '||hi 123||\n'
            '=========='
        )
        self.assertEqual(result, expected)

    def test_multi(self):
        
        text = (
            '  123 456\n'
            '  foo bar baz\n'
            '  foo bar baz xyz  \n'
            '\n'
            '  foo'
            '\n'
        )
        result = box(text)
        expected = (
            '=======================\n'
            '||  123 456          ||\n'
            '||  foo bar baz      ||\n'
            '||  foo bar baz xyz  ||\n'
            '||                   ||\n'
            '||  foo              ||\n'
            '||                   ||\n'
            '======================='
        )
        self.assertEqual(result, expected)

    ## dev-docs: Since the `box` dependency `_paragraph_width` has already been tested to handle `Colored` text, the test below is not needed.
    # def test_colored_text(self):
    #     pass


if __name__ == '__main__':
    unittest.main()



"""
Most of the functions here are standalone
"""
import os as _os
import re as _re
from typing import (
    List as _List,
    Tuple as _Tuple,
    Optional as _Optional,
    Union as _Union,
)


def same_ext_for_all_dir_files(dir_path:str, extension:str) -> bool:
    """
    Check whether all files in `dir_path` have the same `extension`.
    Return `True` if they are all the same, and `False` if not.

    ### Conditions
    - All items in folder `dir_path` are files only; no subfolders.

    ### Params
    - `dir_path`: absolute path to the directory
    - `extension`: file type (including the dot), example: `'.txt'`, `'.py'`.

    ### Exceptions
    - `NotADirectoryError`: if `dir_path` is not a folder
    - `ValueError`: if `extension` doesn't match the regex
    - `AssertionError`: if an item in the folder is not a file

    ### Docs
    - Will return `True` when `dir_path` is empty
    - Will ignore case: `.json` matches `.JSON`, `.TxT` matches `.Txt`, and so on.
    """
    if not _os.path.isdir(dir_path): raise NotADirectoryError(f'Not a dir: {repr(dir_path)}.')
    if not _re.match(r'^\.\w+$', extension): raise ValueError(f'Invalid extension: {repr(extension)}.')
    for file in _os.listdir(dir_path):
        pth = _os.path.join(dir_path, file)
        if not _os.path.isfile(pth): raise AssertionError(f'Not a file: {repr(pth)}.')
        if not file.lower().endswith(extension.lower()): return False
    return True


def list_dir(dir_path:str, /) -> _List[_Tuple[str, str]]:
    """
    The extended version of `os.listdir`.

    @param `dir_path`: absolute path
    @returns: List of pairs of item names and the items' absolute paths,
              e.g., `[(file_name, file_abspath), (subdir_name, subdir_abspath), ...]`

    ### Exceptions
    - `NotADirectoryError`: if `dir_path` not a folder

    ### Demo
    ```
    for file_name, file_path in list_dir('/dir/abs/path'):
        pass
    for item, abspth in list_dir('/dir/abs/path'):
        pass
    for name, pth in list_dir('/dir/abs/path'):  # my favorite  ~Nicholas
        pass
    ```
    """
    if not _os.path.isdir(dir_path): raise NotADirectoryError(f'Not a dir: {repr(dir_path)}.')
    out = []
    for name in _os.listdir(dir_path):
        pth = _os.path.join(dir_path, name)
        out.append((name, pth))
    return out


def remove_all_specific_files_in(dir_path:str, file_pattern:str, recursive:bool=False) -> _List[str]:
    """
    Delete certain files within the `dir_path` folder.

    ---

    ## Params
    - `dir_path`: Absolute path to the folder
    - `file_pattern`: Regex pattern to match the files that you want to delete.
    - `recursive`: If `False`, just delete the files inside `dir_path`.

    ## Returns
    - List of deleted files' absolute paths.

    ## Docs
    - Please review the code for this function before using it.
    """
    deleted = []
    def run(pth):
        for stuff in _os.listdir(pth):
            stuff_pth = _os.path.join(pth, stuff)
            if _os.path.isdir(stuff_pth):
                if recursive: run(stuff_pth)
            else:
                if _re.match(file_pattern, stuff):
                    _os.remove(stuff_pth)
                    deleted.append(stuff_pth)
    run(dir_path)
    return deleted


def norm_pth(__pth:str, /, lowercasing:bool=False) -> str:
    r"""
    Normalizing path.

    ## Params
    - `__pth`: The path
    - `lowercasing`: normalize the case (into lowercase)

    ## Examples
    >>> norm_pth(r'/path/dir////file')  # \path\dir\file
    >>> norm_pth(r'/Path/dir/\/\/\file.TXT/', True)  # \path\dir\file.txt
    """
    if lowercasing: __pth = _os.path.normcase(__pth)
    return _os.path.normpath(__pth)


def definitely_a_dir(__pth:str, /) -> None:
    r"""
    Making sure the given path `__pth` is a directory and it exists

    ## Exceptions
    - `NotADirectoryError`: if not a dir

    ## Demo
    >>> normalized_pth = norm_pth(r'/foo//Bar/\\/\//myDir')
    >>> definitely_a_dir(normalized_pth)
    """
    if not _os.path.isdir(__pth): raise NotADirectoryError(f'Not a dir: {repr(__pth)}.')


def dont_worry_the_path_ends_with(__pth:str, /, suffixes:_Union[str, _List[str], _Tuple[str, ...]], ignore_case:bool=True) -> None:
    """
    Making sure the given path `__pth` ends with the given `suffixes`,
    or if `__pth` is a file path, it guarantees the file has the correct expected extension(s).

    ## Params
    - `__pth`: the path
    - `suffixes`: single or multiple suffixes
    - `ignore_case`: if `False`, ".jpg" is not the same as ".JPG"

    ## Exceptions
    - `ValueError`: if `suffixes` is an empty string
    - `AssertionError`: if the ending does not match the expected `suffixes`

    ## Examples
    - `dont_worry_the_path_ends_with('init_log.txt', '_log.txt')  # okay`
    - `dont_worry_the_path_ends_with('testfile.txt', '_log.txt')  # no`
    - `dont_worry_the_path_ends_with('file.txt', ('.txt', '.log'))  # okay`
    - `dont_worry_the_path_ends_with('file.log', ('.txt', '.log'))  # okay`
    - `dont_worry_the_path_ends_with('file.cpp', ('.txt', '.log'))  # no`
    - `dont_worry_the_path_ends_with('x.foo', ['.FOO', '.BAR'])  # okay`
    - `dont_worry_the_path_ends_with('x.FOO', ['.foo', '.bar'])  # okay`
    - `dont_worry_the_path_ends_with('x.Foo', '.foo', True)   # okay`
    - `dont_worry_the_path_ends_with('x.Foo', '.foo', False)  # no`

    ## Demo
    >>> normalized_pth = norm_pth(r'/foo//Bar/\\/\//file.TXt')
    >>> dont_worry_the_path_ends_with(normalized_pth, '.txt')
    """
    if suffixes == '': raise ValueError("`suffixes` shouldn't be an empty string.")
    
    if ignore_case:
        p_ = __pth.lower()
        if type(suffixes) is str:
            s_ = suffixes.lower()
        else:
            s_ = tuple([s.lower() for s in suffixes])
    else:
        p_ = __pth
        if type(suffixes) is str:
            s_ = suffixes
        else:
            s_ = tuple([s for s in suffixes])
    
    if not p_.endswith(s_):
        raise AssertionError(f'Invalid suffixes: [expected: {repr(suffixes)}] [got: {repr(__pth)}]')


def definitely_a_file(__path:str, /) -> None:
    """
    Making sure the given path is a file that also exists

    ## Exceptions
    - `FileNotFoundError`: if not a file path

    ## Demo
    >>> normalized_pth = norm_pth(r'/foo//Bar/\\/\//file.Txt', True)
    >>> dont_worry_the_path_ends_with(normalized_pth, '.txt')
    >>> definitely_a_file(normalized_pth)
    """
    if not _os.path.isfile(__path): raise FileNotFoundError(f"Not a file: {repr(__path)}.")


def this_path_must_not_exist(__path:str, /) -> None:
    """
    Make sure the given path is not a file, directory, etc.
    This is created to ensure that when performing writing operations, nothing will be overwritten.

    ## Exceptions
    - `AssertionError`: if already exists

    ## Demo
    >>> path_normalized = norm_pth(r'/foo//Bar/\\/\//file.Txt', True)
    >>> dont_worry_the_path_ends_with(path_normalized, '.txt')
    >>> this_path_must_not_exist(path_normalized)
    >>> # rest of the script
    """
    if _os.path.exists(__path): raise AssertionError(f"Already exists: {repr(__path)}.")



import TurboTest as tt
from . import time_quick


def Awesome():

    # print('TT testing: Awesome')

    result = time_quick(0)
    expected = '0m0.0s'
    tt.both_are_equal(result, expected)
    
    result = time_quick(-1)
    expected = '0m1.0s'
    tt.both_are_equal(result, expected)
    
    result = time_quick(1)
    expected = '0m1.0s'
    tt.both_are_equal(result, expected)
    
    result = time_quick(0.1)
    expected = '0m0.1s'
    tt.both_are_equal(result, expected)
    
    result = time_quick(2.13)
    expected = '0m2.1s'
    tt.both_are_equal(result, expected)
    
    result = time_quick(61.1)
    expected = '1m1.1s'
    tt.both_are_equal(result, expected)
    
    result = time_quick(3600)
    expected = '60m0.0s'
    tt.both_are_equal(result, expected)


## TT Testing purposes
# print(locals())
# exec('from kit.readable.time_quick import time_quick')
# print(locals())



## dev-docs: this is the next version of mykit.kit.time.get_dur (this will be deprecated soon)


## TODO: finish this soon
# def time_millennia(
#     __secs: float,
#     /,
#     short: bool = True,
#     include_ms: bool = False,
#     days_in_month: int = 30
# ) -> str:
#     months, _r = divmod(__secs, 3600*24*days_in_month)
#     days, _r = divmod(_r, 3600*24)
#     hours, _r = divmod(_r, 3600)
#     minutes, seconds = divmod(_r, 60)

#     months = int(months)
#     days = int(days)
#     hours = int(hours)
#     minutes = int(minutes)

#     seconds = round(seconds, 3)  # round to milliseconds
#     if not include_ms:
#         seconds = round(seconds)

#     parts = []
#     if short:
#         if months > 0:
#             if months == 1:
#                 parts.append('1 month')
#             else:
#                 parts.append(f'{months} months')
#         if days > 0:
#             if days == 1:
#                 parts.append('1 day')
#             else:
#                 parts.append(f'{days} days')
#         if hours > 0:
#             if hours == 1:
#                 parts.append('1 hr')
#             else:
#                 parts.append(f'{hours} hrs')
#         if minutes > 0:
#             if minutes == 1:
#                 parts.append('1 min')
#             else:
#                 parts.append(f'{minutes} mins')
#         if seconds == 0:
#             if parts == []:
#                 parts.append('0 sec')
#         elif seconds == 1:
#             parts.append('1 sec')
#         else:
#             parts.append(f'{seconds} secs')
#     else:
#         parts.extend((
#             f'{months} months',
#             f'{days} days',
#             f'{hours} hrs',
#             f'{minutes} mins',
#             f'{seconds} secs'
#         ))

#     return ' '.join(parts)




## Birthdate: Oct 6, 2023
def lcut(input_str:str, unwanted:str, check:bool=False) -> str:
    """
    left-cut (lcut). Trim the beginning characters `unwanted` from the given `input_str`.

    ## Params
    - `input_str`: the input string
    - `unwanted`: the characters to be trimmed from the beginning (see examples below to learn more)
    - `check`: if `True`, raise an exception (see below) if `input_str` does not start with `unwanted`

    ## Exceptions
    - `ValueError` if `unwanted` is an empty string
    - `AssertionError` if `input_str` doesn't start with `unwanted` when `check` is `True`

    ## Examples
    - `lcut('121foo', '12')` -> `'1foo'`
    - `lcut('123 hi mom 123', '123 ')` -> `'hi mom 123'`
    - `lcut('hi mom', 'cool')` -> `'hi mom'`

    ## Docs
    - the reason using `lcut` and not `lstrip` is because `'123123 hi mom 123'.lstrip('123')` will be `' hi mom 123'`
    """

    ## Check
    if unwanted == '': raise ValueError("lcut: `unwanted` shouldn't be an empty string.")

    if input_str.startswith(unwanted):
        return input_str[len(unwanted):]
    else:
        if check:
            raise AssertionError("lcut: the given `input_str` does not start with `unwanted`.")
        return input_str



import sys as _sys


## Birthdate: Oct 4, 2023
def get_py_ver() -> str:
    """Get the current Python version in the expected 'Major.minor.patch' format"""
    info = _sys.version_info
    M = info.major  # Major
    m = info.minor  # Minor
    p = info.micro  # Patch
    return f'{M}.{m}.{p}'



import unittest

from mykit.kit.color import (
    interpolate_color,
    getgray,
    rgb_to_hex,
    hexa_to_hex,
    interpolate_with_black,
    hex_to_rgb,
    colored_len, Colored, Hex
)


class Test__Color(unittest.TestCase):

    def test_interpolate_color(self):
        
        ## regular cases

        result = interpolate_color('#ff0000', '#0000ff', 0.0)
        self.assertEqual(result, '#ff0000')

        result = interpolate_color('#ff0000', '#0000ff', 0.5)
        self.assertEqual(result, '#800080')

        result = interpolate_color('#ff0000', '#0000ff', 1.0)
        self.assertEqual(result, '#0000ff')


        ## extreme cases

        result = interpolate_color('#000000', '#ffffff', 0.0)
        self.assertEqual(result, '#000000')

        result = interpolate_color('#000000', '#ffffff', 1.0)
        self.assertEqual(result, '#ffffff')

        result = interpolate_color('#000000', '#000000', 0.5)
        self.assertEqual(result, '#000000')

        result = interpolate_color('#ffffff', '#ffffff', 0.8)
        self.assertEqual(result, '#ffffff')
    
    def test_getgray(self):

        result = getgray(0.5)
        self.assertEqual(result, '#808080')

        result = getgray(0)
        self.assertEqual(result, '#000000')

        result = getgray(1, max_lum=200)
        self.assertEqual(result, '#c8c8c8')

        result = getgray(0.75, max_lum=100)
        self.assertEqual(result, '#4b4b4b')

    def test_rgb_to_hex(self):

        ## black color (all channels set to 0)
        result = rgb_to_hex(0, 0, 0)
        self.assertEqual(result, '#000000')

        ## white color (all channels set to 255)
        result = rgb_to_hex(255, 255, 255)
        self.assertEqual(result, '#ffffff')

        ## red color (only red channel set to 255)
        result = rgb_to_hex(255, 0, 0)
        self.assertEqual(result, '#ff0000')

        ## green color (only green channel set to 255)
        result = rgb_to_hex(0, 255, 0)
        self.assertEqual(result, '#00ff00')

        ## blue color (only blue channel set to 255)
        result = rgb_to_hex(0, 0, 255)
        self.assertEqual(result, '#0000ff')

        ## custom color (random channel values)
        result = rgb_to_hex(100, 150, 200)
        self.assertEqual(result, '#6496c8')

        ## boundary values (minimum and maximum channel values)
        result = rgb_to_hex(0, 255, 127)
        self.assertEqual(result, '#00ff7f')

        ## negative channel values
        result = rgb_to_hex(-10, 0, 255)
        self.assertEqual(result, '#-a00ff')

        ## large channel values
        result = rgb_to_hex(1000, 500, 2550)
        self.assertEqual(result, '#3e81f49f6')
    
    def test_hexa_to_hex(self):

        result = hexa_to_hex('#ffffff', 0, '#000000')
        self.assertEqual(result, '#000000')

        result = hexa_to_hex('#ffffff', 0.5, '#000000')
        self.assertEqual(result, '#808080')

        result = hexa_to_hex('#ffffff', 1, '#000000')
        self.assertEqual(result, '#ffffff')


        result = hexa_to_hex('#abcabc', 0, '#abcabc')
        self.assertEqual(result, '#abcabc')

        result = hexa_to_hex('#abcabc', 0.5, '#abcabc')
        self.assertEqual(result, '#abcabc')

        result = hexa_to_hex('#abcabc', 1, '#abcabc')
        self.assertEqual(result, '#abcabc')


        result = hexa_to_hex('#112233', 0.25, '#aabbcc')
        self.assertEqual(result, '#8495a6')

        result = hexa_to_hex('#ccddee', 0.345, '#123456')
        self.assertEqual(result, '#526e8a')
    
    def test_interpolate_with_black(self):

        result = interpolate_with_black('#ffffff', 0)
        self.assertEqual(result, '#000000')

        result = interpolate_with_black('#ffffff', 0.5)
        self.assertEqual(result, '#808080')

        result = interpolate_with_black('#ffffff', 1)
        self.assertEqual(result, '#ffffff')


        result = interpolate_with_black('#131313', 0.8)
        self.assertEqual(result, '#0f0f0f')

        result = interpolate_with_black('#abc123', 0.234)
        self.assertEqual(result, '#282d08')


class Test__hex_to_rgb(unittest.TestCase):

    def test(self):

        result = hex_to_rgb('#000000')
        expected = (0, 0, 0)
        self.assertEqual(result, expected)

        result = hex_to_rgb('#ffffff')
        expected = (255, 255, 255)
        self.assertEqual(result, expected)

        result = hex_to_rgb('#ff0000')
        expected = (255, 0, 0)
        self.assertEqual(result, expected)

        result = hex_to_rgb('#abcdef')
        expected = (171, 205, 239)
        self.assertEqual(result, expected)


class Test__colored_len(unittest.TestCase):

    def test(self):

        text = 'hi'
        result = colored_len(text)
        self.assertEqual(result, 2)

        text = Colored('hi')
        result = colored_len(text)
        self.assertEqual(result, 2)

        text = Colored('')
        result = colored_len(text)
        self.assertEqual(result, 0)

        text = Colored('12345', Hex.RED, Hex.BLACK)
        result = colored_len(text)
        self.assertEqual(result, 5)

        text = '123' + Colored('45', Hex.RED, Hex.BLACK) + '67'
        result = colored_len(text)
        self.assertEqual(result, 7)

        text = '123' + Colored('45', Hex.RED, Hex.BLACK) + '67' + Colored('89')
        result = colored_len(text)
        self.assertEqual(result, 9)

        text = Colored('12' + Colored('345', Hex.GREEN))
        result = colored_len(text)
        self.assertEqual(result, 5)

        text = Colored('12' + Colored('345', Hex.GREEN, Hex.BLACK)) + '67'
        result = colored_len(text)
        self.assertEqual(result, 7)


if __name__ == '__main__':
    unittest.main()



from typing import (
    List as _List,
    Tuple as _Tuple,
)

from mykit.app import App as _App
from mykit.app.complex.plot import Plot as _Plot


def plot(
    points: _List[_Tuple[float, float]],
    xspan = 0.80,
    yspan = 0.65,
    cfg: dict = {}
):
    """`cfg`: extra configurations"""

    app = _App()

    WIDTH = app.MON_WIDTH*xspan
    HEIGHT = app.MON_HEIGHT*yspan
    TL_X = (app.MON_WIDTH - WIDTH)/2
    TL_Y = (app.MON_HEIGHT - HEIGHT)/2
    _Plot(points, width=WIDTH, height=HEIGHT, tl_x=TL_X, tl_y=TL_Y, **cfg)

    app.run()


import unittest

from mykit.kit.cli_parser.single_simple import SingleSimple

from mykit.kit.color import Colored, Hex
from mykit.kit.stream_capture import StreamCapture


class Test__SingleSimple(unittest.TestCase):

    def setUp(self):
        self.parser = SingleSimple('foo', '123', 'foo bar baz')
        self.parser.add('c1', lambda:print('c1!'), 'test desc')
        self.parser.add('c2345', lambda:print('c2!'), 'test desc 123')

    def test_help_msg(self):
        
        with StreamCapture() as captured:
            self.parser._run_inner(['...'], True)
        
        result = captured.value
        expected = (
            '\n'
            "================================================\n"
            "||                                            ||\n"
            "|| Commands:                                  ||\n"
          # "||   foo       -> show this message then exit ||\n"
          # "||   foo c1    -> test desc                   ||\n"
          # "||   foo c2345 -> test desc 123               ||\n"
            f"||   {Colored('foo c1', Hex.CORN)}    -> {Colored('test desc', Hex.CORN)}                   ||\n"
            f"||   {Colored('foo c2345', Hex.CORN)} -> {Colored('test desc 123', Hex.CORN)}               ||\n"
            f"||   {Colored('foo', Hex.CORN)}       -> {Colored('show this message then exit', Hex.CORN)} ||\n"
            "||                                            ||\n"
            "|| Info:                                      ||\n"
            "||   software  : foo                          ||\n"
            "||   version   : 123                          ||\n"
            "||   repository: foo bar baz                  ||\n"
            "||                                            ||\n"
            "================================================"
            '\n'  # effect from the `print` function
        )
        self.assertEqual(result, expected)

    def test_invalid_input_command(self):

        with StreamCapture() as captured:
            self.parser._run_inner(['...', 'c3'], True)
        
        result = captured.value
        expected = "Unknown commands 'c3', run `foo` for help.\n"
        self.assertEqual(result, expected)
    
    def test_run_the_functions(self):
        
        with StreamCapture() as captured:
            self.parser._run_inner(['...', 'c1'], True)
        self.assertEqual(captured.value, 'c1!\n')

        with StreamCapture() as captured:
            self.parser._run_inner(['...', 'c2345'], True)
        self.assertEqual(captured.value, 'c2!\n')


if __name__ == '__main__':
    unittest.main()



import unittest

import os
import random
import tempfile
import shutil

from mykit.kit.fileops.simple import (
    same_ext_for_all_dir_files, list_dir, remove_all_specific_files_in, definitely_a_dir,
    dont_worry_the_path_ends_with,
    definitely_a_file,
    this_path_must_not_exist,
)


class Test__same_ext_for_all_dir_files(unittest.TestCase):

    def test_core_I(self):  # Empty dir
        dir = tempfile.mkdtemp()
        result = same_ext_for_all_dir_files(dir, '.foo')
        self.assertEqual(result, True)

    def test_core_II(self):  # All files are the same type
        
        dir = tempfile.mkdtemp()
        n = random.randint(1, 10)
        for i in range(n): open(os.path.join(dir, f'file_{i}.TXT'), 'w').close()

        self.assertEqual(len(os.listdir(dir)), n)  # Debugging purposes

        result = same_ext_for_all_dir_files(dir, '.txt')
        self.assertEqual(result, True)

    def test_core_III(self):  # Files are of different types
        
        dir = tempfile.mkdtemp()
        n = random.randint(1, 10)
        for i in range(n): open(os.path.join(dir, f'file_{i}.txT'), 'w').close()
        for i in range(n): open(os.path.join(dir, f'file_{i}.mdx'), 'w').close()

        result = same_ext_for_all_dir_files(dir, '.tXt')
        self.assertEqual(result, False)

    def test_not_a_dir(self):

        ## Test I

        with self.assertRaises(NotADirectoryError) as ctx: same_ext_for_all_dir_files('foo', '.foo')
        self.assertEqual(str(ctx.exception), "Not a dir: 'foo'.")

        ## Test II
        
        dir = tempfile.mkdtemp()
        pth = os.path.join(dir, 'file.txt')
        open(pth, 'w').close()

        self.assertEqual(os.path.isfile(pth), True)  # Debugging purposes
        
        with self.assertRaises(NotADirectoryError) as ctx: same_ext_for_all_dir_files(pth, '.foo')
        self.assertEqual(str(ctx.exception), f'Not a dir: {repr(pth)}.')

    def test_extension_validity(self):

        dir = tempfile.mkdtemp()

        ## Passes

        same_ext_for_all_dir_files(dir, '.foo')
        same_ext_for_all_dir_files(dir, '.123foo')
        same_ext_for_all_dir_files(dir, '.foo_bar_123')

        ## Fails

        with self.assertRaises(ValueError) as ctx: same_ext_for_all_dir_files(dir, '')
        self.assertEqual(str(ctx.exception), "Invalid extension: ''.")

        with self.assertRaises(ValueError) as ctx: same_ext_for_all_dir_files(dir, 'txt')
        self.assertEqual(str(ctx.exception), "Invalid extension: 'txt'.")

        with self.assertRaises(ValueError) as ctx: same_ext_for_all_dir_files(dir, '.txt ')
        self.assertEqual(str(ctx.exception), "Invalid extension: '.txt '.")

        with self.assertRaises(ValueError) as ctx: same_ext_for_all_dir_files(dir, '.txt+')
        self.assertEqual(str(ctx.exception), "Invalid extension: '.txt+'.")
        
    def test_item_is_not_file(self):

        dir = tempfile.mkdtemp()

        ## Dummy data
        for i in range(3): open(os.path.join(dir, f'file_{i}.py'), 'w').close()
        pth = os.path.join(dir, 'subdir')
        os.mkdir(pth)

        with self.assertRaises(AssertionError) as ctx: same_ext_for_all_dir_files(dir, '.py')
        self.assertEqual(str(ctx.exception), f'Not a file: {repr(pth)}.')


class Test__list_dir(unittest.TestCase):

    def test_core_I(self):  # Empty dir
        dir = tempfile.mkdtemp()
        result = list_dir(dir)
        expected = []
        self.assertEqual(result, expected)

    def test_core_II(self):  # normal
        
        dir = tempfile.mkdtemp()
        name = 'file.py'
        pth = os.path.join(dir, name)
        open(pth, 'w').close()

        result = list_dir(dir)
        expected = [(name, pth)]
        self.assertEqual(result, expected)

    def test_core_III(self):  # normal
        
        dir = tempfile.mkdtemp()
        for i in range(3): open(os.path.join(dir, f'file_{i}.txt'), 'w').close()

        result = sorted(list_dir(dir))  # Sorting to double-check expected behavior
        expected = [
            ('file_0.txt', os.path.join(dir, 'file_0.txt')),
            ('file_1.txt', os.path.join(dir, 'file_1.txt')),
            ('file_2.txt', os.path.join(dir, 'file_2.txt')),
        ]
        self.assertEqual(result, expected)

    def test_core_IV(self):  # normal
        
        dir = tempfile.mkdtemp()
        for i in range(3): open(os.path.join(dir, f'file_{i}.txt'), 'w').close()
        for i in range(3): os.mkdir(os.path.join(dir, f'subdir_{i}'))

        self.assertEqual(len(os.listdir(dir)), 6)  # Debugging purposes

        result = sorted(list_dir(dir))  # Sorting to double-check expected behavior
        expected = [
            ('file_0.txt', os.path.join(dir, 'file_0.txt')),
            ('file_1.txt', os.path.join(dir, 'file_1.txt')),
            ('file_2.txt', os.path.join(dir, 'file_2.txt')),
            ('subdir_0', os.path.join(dir, 'subdir_0')),
            ('subdir_1', os.path.join(dir, 'subdir_1')),
            ('subdir_2', os.path.join(dir, 'subdir_2')),
        ]
        self.assertEqual(result, expected)


class Test__remove_all_specific_files_in(unittest.TestCase):

    def setUp(self):
        self.dir = tempfile.mkdtemp()

        ## Dummy file

        open(os.path.join(self.dir, 'foo.txt'), 'w').close()
        open(os.path.join(self.dir, 'bar.txt'), 'w').close()
        open(os.path.join(self.dir, 'baz.txt'), 'w').close()

        open(os.path.join(self.dir, 'abc.py'), 'w').close()
        open(os.path.join(self.dir, 'xyz.py'), 'w').close()

        ## Subdir

        os.mkdir(os.path.join(self.dir, 'subdir'))
        open(os.path.join(self.dir, 'subdir', 'test1.txt'), 'w').close()
        open(os.path.join(self.dir, 'subdir', 'test2.txt'), 'w').close()
        open(os.path.join(self.dir, 'subdir', 'foobar.js'), 'w').close()

        ## Deeper subdir

        os.mkdir(os.path.join(self.dir, 'subdir', 'subdir2'))
        open(os.path.join(self.dir, 'subdir', 'subdir2', 'f1.txt'), 'w').close()
        open(os.path.join(self.dir, 'subdir', 'subdir2', 'f2.txt'), 'w').close()
        open(os.path.join(self.dir, 'subdir', 'subdir2', 'b.ts'), 'w').close()

        ## Validate

        self.assertEqual(sorted(os.listdir(self.dir)), ['abc.py', 'bar.txt', 'baz.txt', 'foo.txt', 'subdir', 'xyz.py'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir'))), ['foobar.js', 'subdir2', 'test1.txt', 'test2.txt'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir', 'subdir2'))), ['b.ts', 'f1.txt', 'f2.txt'])

    def tearDown(self):
        shutil.rmtree(self.dir)

    def test_delete_every_file(self):
        deleted = remove_all_specific_files_in(self.dir, r'.*', True)
        self.assertEqual(sorted(deleted), [
            os.path.join(self.dir, 'abc.py'),
            os.path.join(self.dir, 'bar.txt'),
            os.path.join(self.dir, 'baz.txt'),
            os.path.join(self.dir, 'foo.txt'),
            os.path.join(self.dir, 'subdir', 'foobar.js'),
            os.path.join(self.dir, 'subdir', 'subdir2', 'b.ts'),
            os.path.join(self.dir, 'subdir', 'subdir2', 'f1.txt'),
            os.path.join(self.dir, 'subdir', 'subdir2', 'f2.txt'),
            os.path.join(self.dir, 'subdir', 'test1.txt'),
            os.path.join(self.dir, 'subdir', 'test2.txt'),
            os.path.join(self.dir, 'xyz.py'),
        ])
        self.assertEqual(sorted(os.listdir(self.dir)), ['subdir'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir'))), ['subdir2'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir', 'subdir2'))), [])

    def test_delete_based_on_extension(self):
        remove_all_specific_files_in(self.dir, r'.*?\.txt$', True)
        self.assertEqual(sorted(os.listdir(self.dir)), ['abc.py', 'subdir', 'xyz.py'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir'))), ['foobar.js', 'subdir2'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir', 'subdir2'))), ['b.ts'])

    def test_delete_based_on_prefix(self):
        remove_all_specific_files_in(self.dir, r'^f.*', True)
        self.assertEqual(sorted(os.listdir(self.dir)), ['abc.py', 'bar.txt', 'baz.txt', 'subdir', 'xyz.py'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir'))), ['subdir2', 'test1.txt', 'test2.txt'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir', 'subdir2'))), ['b.ts'])

    def test_delete_no_match(self):
        remove_all_specific_files_in(self.dir, r'999', True)
        self.assertEqual(sorted(os.listdir(self.dir)), ['abc.py', 'bar.txt', 'baz.txt', 'foo.txt', 'subdir', 'xyz.py'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir'))), ['foobar.js', 'subdir2', 'test1.txt', 'test2.txt'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir', 'subdir2'))), ['b.ts', 'f1.txt', 'f2.txt'])

    def test_delete_no_recursive(self):
        deleted = remove_all_specific_files_in(self.dir, r'.*?\.py$', False)
        self.assertEqual(sorted(deleted), [
            os.path.join(self.dir, 'abc.py'),
            os.path.join(self.dir, 'xyz.py'),
        ])
        self.assertEqual(sorted(os.listdir(self.dir)), ['bar.txt', 'baz.txt', 'foo.txt', 'subdir'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir'))), ['foobar.js', 'subdir2', 'test1.txt', 'test2.txt'])
        self.assertEqual(sorted(os.listdir(os.path.join(self.dir, 'subdir', 'subdir2'))), ['b.ts', 'f1.txt', 'f2.txt'])


class Test__definitely_a_dir(unittest.TestCase):

    def test(self):

        ## Pass

        p = tempfile.mkdtemp()
        pth = os.path.join(p, 'test_dir')
        os.mkdir(pth)
        definitely_a_dir(pth)

        ## Fail

        pth = os.path.join(p, 'test_dir2')
        with self.assertRaises(NotADirectoryError) as ctx: definitely_a_dir(pth)
        self.assertEqual(str(ctx.exception), f"Not a dir: {repr(pth)}.")


class Test__dont_worry_the_path_ends_with(unittest.TestCase):

    def test_empty_string_suffixes(self):
        with self.assertRaises(ValueError) as ctx: dont_worry_the_path_ends_with('/path/foo', '')
        self.assertEqual(str(ctx.exception), "`suffixes` shouldn't be an empty string.")

    def test_core(self):
        
        dont_worry_the_path_ends_with('init_log.txt', '_log.txt')

        with self.assertRaises(AssertionError) as ctx: dont_worry_the_path_ends_with('testfile.txt', '_log.txt')
        self.assertEqual(str(ctx.exception), "Invalid suffixes: [expected: '_log.txt'] [got: 'testfile.txt']")

        dont_worry_the_path_ends_with('file.txt', ('.txt', '.log'))
        dont_worry_the_path_ends_with('file.log', ('.txt', '.log'))

        with self.assertRaises(AssertionError) as ctx: dont_worry_the_path_ends_with('file.cpp', ('.txt', '.log'))
        self.assertEqual(str(ctx.exception), "Invalid suffixes: [expected: ('.txt', '.log')] [got: 'file.cpp']")
        
        dont_worry_the_path_ends_with('x.foo', ['.FOO', '.BAR'])
        dont_worry_the_path_ends_with('x.FOO', ['.foo', '.bar'])
        dont_worry_the_path_ends_with('x.Foo', '.foo', True)

        with self.assertRaises(AssertionError) as ctx: dont_worry_the_path_ends_with('x.Foo', '.foo', False)
        self.assertEqual(str(ctx.exception), "Invalid suffixes: [expected: '.foo'] [got: 'x.Foo']")


class Test__definitely_a_file(unittest.TestCase):

    def test(self):
        
        p = tempfile.mkdtemp()
        
        ## Pass

        pth = os.path.join(p, 'test_file.txt')
        open(pth, 'w').close()
        definitely_a_file(pth)

        ## Fail

        pth = os.path.join(p, 'test_file2.txt')
        with self.assertRaises(FileNotFoundError) as ctx: definitely_a_file(pth)
        self.assertEqual(str(ctx.exception), f"Not a file: {repr(pth)}.")


class Test__this_path_must_not_exist(unittest.TestCase):

    def test(self):
        
        p = tempfile.mkdtemp()
        

        ## Pass
        ## vvvv

        ## file
        pth = os.path.join(p, 'test_file.txt')
        self.assertEqual(os.path.isfile(pth), False)
        this_path_must_not_exist(pth)

        ## dir
        pth = os.path.join(p, 'test_dir.txt')
        self.assertEqual(os.path.isdir(pth), False)
        this_path_must_not_exist(pth)


        ## Fail
        ## vvvv

        ## file
        pth = os.path.join(p, 'test_file2.txt')
        open(pth, 'w').close()
        self.assertEqual(os.path.isfile(pth), True)
        with self.assertRaises(AssertionError) as ctx: this_path_must_not_exist(pth)
        self.assertEqual(str(ctx.exception), f"Already exists: {repr(pth)}.")

        ## dir
        pth = os.path.join(p, 'test_dir2.txt')
        os.mkdir(pth)
        self.assertEqual(os.path.isdir(pth), True)
        with self.assertRaises(AssertionError) as ctx: this_path_must_not_exist(pth)
        self.assertEqual(str(ctx.exception), f"Already exists: {repr(pth)}.")


if __name__ == '__main__':
    unittest.main()



import io as _io
import sys as _sys


class StreamCapture:
    """Redirects and captures stream content"""

    def __init__(self, fd:int=1):
        r"""
        Initialize the context manager.

        ---

        ## Params
        - `fd`: file descriptors: `0` for stdin, `1` for stdout, and `2` is for stderr.

        ## Demo
        >>> with StreamCapture() as captured:
        >>>     print('foo\n\nbar')
        >>>     print('baz')
        >>> print(repr(captured.value))  # 'foo\n\nbar\nbaz\n'
        """
        self.fd = fd

        if fd in [0, 2]: raise NotImplementedError  # TODO: Remove this once implemented

        if fd not in [0, 1, 2]: raise ValueError(f'Invalid `fd` value: {repr(fd)}.')

        self.holder = _io.StringIO()

        ## <Runtime>
        self.old = None
        ## </Runtime>

    def __enter__(self):

        ## Save
        if self.fd == 0: self.old = _sys.stdin  # Hasn't tested
        if self.fd == 1: self.old = _sys.stdout
        if self.fd == 2: self.old = _sys.stderr  # Hasn't tested

        ## Listening
        self.holder = self.holder
        if self.fd == 0: _sys.stdin = self.holder  # Hasn't tested
        if self.fd == 1: _sys.stdout = self.holder
        if self.fd == 2: _sys.stderr = self.holder  # Hasn't tested

        return self

    def __exit__(self, exc_type, exc_value, traceback):

        ## Restore
        if self.fd == 0: _sys.stdin = self.old  # Hasn't tested
        if self.fd == 1: _sys.stdout = self.old
        if self.fd == 2: _sys.stderr = self.old  # Hasn't tested

    @property
    def value(self):
        """The captured output as a string"""
        return self.holder.getvalue()


import unittest  # The testing framework

## The ones being tested
from mykit.kit.shell import (
    run
)

## Additional helper modules
import os
import tempfile


class Test__run(unittest.TestCase):

    def test(self):
        cwd = tempfile.mkdtemp()
        self.assertEqual(os.listdir(cwd), [])
        run('git init --quiet', cwd=cwd)
        self.assertEqual(os.listdir(cwd), ['.git'])
        run('mkdir foo', cwd=cwd)
        self.assertEqual(sorted(os.listdir(cwd)), ['.git', 'foo'])


if __name__ == '__main__':
    unittest.main()


"""
This module is a build module for constructing mykit-related software/programs/libraries distribution.

Note, for distribution, this 'inventory' master module shouldn't be included since it has nothing to do with the end user.
"""



import random as _random
import tkinter as _tk
from typing import (
    Callable as _Callable,
    Dict as _Dict,
    List as _List,
    Literal as _Literal,
    Optional as _Optional,
    Tuple as _Tuple,
    Union as _Union
)

from mykit.kit.utils import minmax_normalization as _norm


class _Slider:

    ## <runtime>

    _page: _tk.Canvas = None
    @staticmethod
    def _set_page(page):
        _Slider._page = page

    sliders: _Dict[str, '_Slider'] = {}
    slider_tags: _Dict[str, _List['_Slider']] = {}  # note that the horizontal and vertical sliders store the tags together

    ## </runtime>

    def __init__(
        self,
        min: float = 0,
        max: float = 1,
        step: _Optional[float] = None,
        init: _Optional[float] = None,
        fn: _Optional[_Callable[[], None]] = None,

        x: int = 0,
        y: int = 0,
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'] = 'nw',
        tolerance: float = 0.25,
        
        label: _Optional[str] = None,
        label_fg: str = '#ccc',
        label_font: _Union[str, _Tuple[str, int]] = 'Verdana 9',
        show_label_box: bool = False,
        label_box_color: _Optional[str] = None,
        label_box_width: _Optional[int] = None,
        label_box_height: _Optional[int] = None,
        label_y_shift: int = -15,

        show_value: bool = True,
        value_fg: str = '#ccc',
        value_font: _Union[str, _Tuple[str, int]] = 'Consolas 9',
        value_prefix: str = '',
        value_suffix: str = '',
        show_value_box: bool = False,
        value_box_color: _Optional[str] = None,
        value_box_width: _Optional[int] = None,
        value_box_height: _Optional[int] = None,

        rod_len: int = 200,
        rod_thick: int = 3,
        btn_r: int = 5,
        color_rod_normal: str = '#4b4b4b',
        color_rod_locked: str = '#2d2d2d',
        color_btn_normal: str = '#555555',
        color_btn_locked: str = '#373737',
        color_btn_press: str = '#7d7d7d',
        color_btn_hover: str = '#696969',
        color_btn_bd_normal: str = '#f5f5f5',
        color_btn_bd_locked: str = '#373737',

        locked: bool = False,
        visible: bool = True,

        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ) -> None:
        """
        ## Params
        - `x` and `y` is the position of the `anchor` (not the center of the button)
        - `tolerance`: to determine how closely the cursor needs to be to the slider's step values for it to move to a new value

        ## Docs
        - box color, width, and height should be provided if the box shown
        """

        ## <dependencies check>

        if _Slider._page is None:
            raise AssertionError('Can\'t use widgets before App initialized.')
        
        ## </dependencies check>

        self.min = min
        self.max = max
        self.step = step if step is not None else (max-min)/10
        self.init = init if init is not None else (max-min)/2
        self.fn = fn

        self.x = x
        self.y = y
        self.anchor = anchor
        self.tolerance = tolerance

        self.label = label
        self.label_fg = label_fg
        self.label_font = label_font
        self.show_label_box = show_label_box
        self.label_box_color = label_box_color
        self.label_box_width = label_box_width
        self.label_box_height = label_box_height
        self.label_y_shift = label_y_shift

        self.show_value = show_value
        self.value_fg = value_fg
        self.value_font = value_font
        self.value_prefix = value_prefix
        self.value_suffix = value_suffix
        self.show_value_box = show_value_box
        self.value_box_color = value_box_color
        self.value_box_width = value_box_width
        self.value_box_height = value_box_height

        self.rod_len = rod_len
        self.rod_thick = rod_thick
        self.btn_r = btn_r
        self.color_rod_normal = color_rod_normal
        self.color_rod_locked = color_rod_locked
        self.color_btn_normal = color_btn_normal
        self.color_btn_locked = color_btn_locked
        self.color_btn_press = color_btn_press
        self.color_btn_hover = color_btn_hover
        self.color_btn_bd_normal = color_btn_bd_normal
        self.color_btn_bd_locked = color_btn_bd_locked

        self.locked = locked
        self.visible = visible

        ## <id>

        ## self.id ensures that we can modify a specific instance without affecting the others
        if id is None:
            self.id = str(_random.randint(0, 100_000))
            while self.id in _Slider.sliders:
                self.id = str(_random.randint(0, 100_000))
        else:
            self.id = id
            if self.id in _Slider.sliders:
                raise ValueError(f'The id {repr(id)} is duplicated.')
        
        _Slider.sliders[self.id] = self

        ## </id>


        ## <tags>

        if type(tags) is str:
            self.tags = [tags]
        elif (type(tags) is list) or (type(tags) is tuple) or (tags is None):
            self.tags = tags
        
        if tags is not None:
            for tag in self.tags:
                if tag in _Slider.slider_tags:
                    _Slider.slider_tags[tag].append(self)
                else:
                    _Slider.slider_tags[tag] = [self]

        ## </tags>


        ## runtime

        self.value = self.init
        self.prec = len(str(abs(self.step)).split('.')[1]) if '.' in str(self.step) else 0
        self.pressed = False
        self.hovered = False


    @staticmethod
    def _hover_listener():  # reminder: hover listeners don't need `e: _tk.Event` arg
        for slider in _Slider.sliders.values():
            slider._hover()

    @staticmethod
    def _press_listener(e: _tk.Event):
        for slider in _Slider.sliders.values():
            slider._press()

    @staticmethod
    def _hold_listener(e: _tk.Event):
        for slider in list(_Slider.sliders.values()):
            slider._hold()
    

    def _release(self):
        if self.pressed:
            self.pressed = False
            self._redraw()

    @staticmethod
    def _release_listener(e: _tk.Event):
        for slider in list(_Slider.sliders.values()):
            slider._release()


    def set_lock(self, locked: bool, /) -> None:
        if self.locked is not locked:
            self.locked = locked
            self._redraw()
    
    @staticmethod
    def set_lock_by_id(id: str, locked: bool, /) -> None:
        _Slider.sliders[id].set_lock(locked)

    @staticmethod
    def set_lock_by_tag(tag: str, locked: bool, /) -> None:
        for slider in _Slider.slider_tags[tag]:
            slider.set_lock(locked)
    
    @staticmethod
    def set_lock_all(locked: bool, /) -> None:
        for slider in _Slider.sliders.values():
            slider.set_lock(locked)


    def set_visibility(self, visible: bool, /) -> None:
        if self.visible is not visible:
            self.visible = visible
            self._redraw()
    
    @staticmethod
    def set_visibility_by_id(id: str, visible: bool, /) -> None:
        _Slider.sliders[id].set_visibility(visible)
    
    @staticmethod
    def set_visibility_by_tag(tag: str, visible: bool, /) -> None:
        for slider in _Slider.slider_tags[tag]:
            slider.set_visibility(visible)
    
    @staticmethod
    def set_visibility_all(visible: bool, /) -> None:
        for slider in _Slider.sliders.values():
            slider.set_visibility(visible)


    def set_value(self, value: _Optional[int], /) -> None:
        """if `None` -> default value."""

        if value is None:
            value = self.init

        if value < self.min:
            value = self.min
        if value > self.max:
            value = self.max

        if value != self.value:
            self.value = value
            self._redraw()
    
    @staticmethod
    def set_value_by_id(id: str, value: _Optional[int], /) -> None:
        _Slider.sliders[id].set_value(value)

    @staticmethod
    def set_value_by_tag(tag: str, value: _Optional[int], /) -> None:
        for slider in _Slider.slider_tags[tag]:
            slider.set_value(value)

    @staticmethod
    def set_value_all(value: _Optional[int], /) -> None:
        """To reset the value of all sliders, use `value = None`."""
        for slider in _Slider.sliders.values():
            slider.set_value(value)


    @staticmethod
    def get_value_by_id(id: str, /) -> None:
        return _Slider.sliders[id].value
    

    def get_anchor_loc(
        self,
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'],
        /
    ) -> _Tuple[int, int]:
        """To get the slider-rod's center coordinate, use `anchor='center'`"""

        W = self.width
        H = self.height

        ## get the coordinates for the center (X, Y)
        if self.anchor == 'center':
            X = self.x
            Y = self.y
        elif self.anchor == 'n':
            X = self.x
            Y = self.y + H/2
        elif self.anchor == 'ne':
            X = self.x - W/2
            Y = self.y + H/2
        elif self.anchor == 'e':
            X = self.x - W/2
            Y = self.y
        elif self.anchor == 'se':
            X = self.x - W/2
            Y = self.y - H/2
        elif self.anchor == 's':
            X = self.x
            Y = self.y - H/2
        elif self.anchor == 'sw':
            X = self.x + W/2
            Y = self.y - H/2
        elif self.anchor == 'w':
            X = self.x + W/2
            Y = self.y
        elif self.anchor == 'nw':
            X = self.x + W/2
            Y = self.y + H/2

        ## returning the requested anchor location
        if anchor == 'center':
            return (X, Y)
        elif anchor == 'n':
            return (X, Y-H/2)
        elif anchor == 'ne':
            return (X+W/2, Y-H/2)
        elif anchor == 'e':
            return (X+W/2, Y)
        elif anchor == 'se':
            return (X+W/2, Y+H/2)
        elif anchor == 's':
            return (X, Y+H/2)
        elif anchor == 'sw':
            return (X-W/2, Y+H/2)
        elif anchor == 'w':
            return (X-W/2, Y)
        elif anchor == 'nw':
            return (X-W/2, Y-H/2)
        else:
            raise ValueError(f'Invalid anchor value: {repr(anchor)}')

    @staticmethod
    def get_anchor_loc_by_id(
        id: str,
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'],
        /
    ) -> _Tuple[int, int]:
        return _Slider.sliders[id].get_anchor_loc(anchor)


    def move(
        self,
        x: int,
        y: int,
        /,
        anchor: _Optional[_Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']] = None
    ) -> '_Slider':
        """If `anchor = None`, the current anchor will be used."""
        self.x = x
        self.y = y
        if anchor is not None:
            self.anchor = anchor
        self._redraw()
        return self

    @staticmethod
    def move_by_id(
        id: str,
        x: int,
        y: int,
        /,
        anchor: _Optional[_Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']] = None
    ) -> None:
        _Slider.sliders[id].move(x, y, anchor)


    def align(
        self,
        target: '_Slider',
        anchor: str = 'nw',
        target_anchor: str = 'ne',
        xgap: float = 15,
        ygap: float = 0
    ) -> '_Slider':
        """
        Valid options for `anchor` and `target_anchor` are `['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw']`.
        """

        ## getting the target anchor location
        x, y = target.get_anchor_loc(target_anchor)

        ## shifting
        x += xgap
        y += ygap

        ## moving the label
        self.move(x, y, anchor)

        ## return the instance so that this method can also be used when
        ## creating the instance, like `sld_2 = Slider().align(sld_1)`.
        return self


    def destroy(self) -> None:
        _Slider.sliders.pop(self.id)

        if self.tags is not None:
            for tag in self.tags:
                _Slider.slider_tags[tag].remove(self)
                if _Slider.slider_tags[tag] == []:
                    _Slider.slider_tags.pop(tag)

        _Slider._page.delete(f'Slider_{self.id}')

    @staticmethod
    def destroy_by_id(id: str, /) -> None:
        _Slider.sliders[id].destroy()

    @staticmethod
    def destroy_by_tag(tag: str, /) -> None:
        for slider in list(_Slider.slider_tags[tag]):
            slider.destroy()

    @staticmethod
    def destroy_all() -> None:
        for slider in list(_Slider.sliders.values()):
            slider.destroy()


class Slider(_Slider):

    def __init__(
        self,
        min: float = 0,
        max: float = 1,
        step: _Optional[float] = None,
        init: _Optional[float] = None,
        fn: _Optional[_Callable[[], None]] = None,

        x: int = 0,
        y: int = 0,
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'] = 'nw',
        tolerance: float = 0.25,
        
        label: _Optional[str] = None,
        label_fg: str = '#ccc',
        label_font: _Union[str, _Tuple[str, int]] = 'Verdana 9',
        show_label_box: bool = False,
        label_box_color: _Optional[str] = None,
        label_box_width: _Optional[int] = None,
        label_box_height: _Optional[int] = None,
        label_y_shift: int = -15,

        show_value: bool = True,
        value_fg: str = '#ccc',
        value_font: _Union[str, _Tuple[str, int]] = 'Consolas 9',
        value_prefix: str = '',
        value_suffix: str = '',
        show_value_box: bool = False,
        value_box_color: _Optional[str] = None,
        value_box_width: _Optional[int] = None,
        value_box_height: _Optional[int] = None,
        value_box_x_shift: int = 25,  # extra arg. reminder: this parameter has a different argument name for horizontal and vertical sliders

        rod_len: int = 200,
        rod_thick: int = 3,
        btn_r: int = 5,
        color_rod_normal: str = '#4b4b4b',
        color_rod_locked: str = '#2d2d2d',
        color_btn_normal: str = '#555555',
        color_btn_locked: str = '#373737',
        color_btn_press: str = '#7d7d7d',
        color_btn_hover: str = '#696969',
        color_btn_bd_normal: str = '#f5f5f5',
        color_btn_bd_locked: str = '#373737',

        locked: bool = False,
        visible: bool = True,

        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ) -> None:

        super().__init__(
            min, max, step, init, fn,
            x, y, anchor, tolerance,
            label, label_fg, label_font, show_label_box, label_box_color, label_box_width, label_box_height, label_y_shift,
            show_value, value_fg, value_font, value_prefix, value_suffix, show_value_box, value_box_color, value_box_width, value_box_height,
            rod_len, rod_thick, btn_r, color_rod_normal, color_rod_locked, color_btn_normal, color_btn_locked, color_btn_press, color_btn_hover, color_btn_bd_normal, color_btn_bd_locked,
            locked, visible,
            id, tags
        )

        self.value_box_x_shift = value_box_x_shift

        ## runtime (custom)

        self.width = self.rod_len
        self.height = self.rod_thick

        ## init        
        
        self._redraw()

    def _redraw(self):

        if self.locked:
            color_rod = self.color_rod_locked
            color_btn = self.color_btn_locked
            color_btn_bd = self.color_btn_bd_locked
        elif self.pressed:
            color_rod = self.color_rod_normal
            color_btn = self.color_btn_press
            color_btn_bd = self.color_btn_bd_normal
        elif self.hovered:
            color_rod = self.color_rod_normal
            color_btn = self.color_btn_hover
            color_btn_bd = self.color_btn_bd_normal
        else:
            color_rod = self.color_rod_normal
            color_btn = self.color_btn_normal
            color_btn_bd = self.color_btn_bd_normal

        _Slider._page.delete(f'Slider_{self.id}')

        if self.visible:

            X, Y = self.get_anchor_loc('center')
            w2 = self.width/2
            h2 = self.height/2

            _Slider._page.create_rectangle(
                X-w2, Y-h2,
                X+w2, Y+h2,
                fill=color_rod, width=0, tags=f'Slider_{self.id}'
            )
            _Slider._page.create_oval(
                X-w2+_norm(self.value, self.min, self.max)*self.width-self.btn_r, Y-self.btn_r,
                X-w2+_norm(self.value, self.min, self.max)*self.width+self.btn_r, Y+self.btn_r,
                fill=color_btn, width=1, outline=color_btn_bd, tags=f'Slider_{self.id}'
            )

            if self.show_label_box:
                _Slider._page.create_rectangle(
                    X-self.label_box_width/2, Y-h2+self.label_y_shift-self.label_box_height/2,
                    X+self.label_box_width/2, Y-h2+self.label_y_shift+self.label_box_height/2,
                    fill=self.label_box_color, width=0, tags=f'Slider_{self.id}'
                )
            if self.label is not None:
                _Slider._page.create_text(
                    X, Y-h2+self.label_y_shift,
                    text=self.label, font=self.label_font, fill=self.label_fg, tags=f'Slider_{self.id}'
                )

            if self.show_value_box:
                _Slider._page.create_rectangle(
                    X+w2+self.value_box_x_shift-self.value_box_width/2, Y-self.value_box_height/2,
                    X+w2+self.value_box_x_shift+self.value_box_width/2, Y+self.value_box_height/2,
                    fill=self.value_box_color, width=0, tags=f'Slider_{self.id}'
                )
            if self.show_value:
                _Slider._page.create_text(
                    X+w2+self.value_box_x_shift, Y,
                    text=self.value_prefix + str(self.value) + self.value_suffix,
                    font=self.value_font, fill=self.value_fg, tags=f'Slider_{self.id}'
                )

    def _hover(self):

        X, Y = self.get_anchor_loc('center')

        x = _Slider._page.winfo_pointerx()
        y = _Slider._page.winfo_pointery()

        ## button coordinate
        bx = X - self.width/2 + _norm(self.value, self.min, self.max)*self.width
        by = Y
        br = self.btn_r  # button radius

        ## `True` if the mouse cursor is inside the slider button
        inside = (bx-br <= x <= bx+br) and (by-br <= y <= by+br)

        if inside and (not self.locked) and self.visible and (not self.hovered):
            self.hovered = True
            self._redraw()
        
        elif self.hovered and (not inside):
            self.hovered = False
            self._redraw()

    def _press(self):

        X, Y = self.get_anchor_loc('center')

        x = _Slider._page.winfo_pointerx()
        y = _Slider._page.winfo_pointery()

        bx = X - self.width/2 + _norm(self.value, self.min, self.max)*self.width
        by = Y
        br = self.btn_r

        inside = (bx-br <= x <= bx+br) and (by-br <= y <= by+br)

        if inside and (not self.locked) and self.visible:
            self.pressed = True
            self._redraw()

    def _hold(self):

        if self.pressed:

            X, _ = self.get_anchor_loc('center')

            mousex = _Slider._page.winfo_pointerx()
            value = self.min + ((mousex - (X - self.width/2))/self.width)*(self.max - self.min)

            if self.prec == 0:
                value = int(value)
            else:
                value = round(value, self.prec)

            if abs(value - round(value/self.step)*self.step) < self.tolerance*self.step:

                if value < self.min:
                    value = self.min
                if value > self.max:
                    value = self.max
                
                if value == self.value:
                    return
                self.value = value

                self._redraw()
                if self.fn is not None:
                    self.fn()


class VSlider(_Slider):

    def __init__(
        self,
        min: float = 0,
        max: float = 1,
        step: _Optional[float] = None,
        init: _Optional[float] = None,
        fn: _Optional[_Callable[[], None]] = None,

        x: int = 0,
        y: int = 0,
        anchor: _Literal['center', 'n', 'ne', 'e', 'se', 's', 'sw', 'w', 'nw'] = 'nw',
        tolerance: float = 0.25,
        
        label: _Optional[str] = None,
        label_fg: str = '#ccc',
        label_font: _Union[str, _Tuple[str, int]] = 'Verdana 9',
        show_label_box: bool = False,
        label_box_color: _Optional[str] = None,
        label_box_width: _Optional[int] = None,
        label_box_height: _Optional[int] = None,
        label_y_shift: int = -15,

        show_value: bool = True,
        value_fg: str = '#ccc',
        value_font: _Union[str, _Tuple[str, int]] = 'Consolas 9',
        value_prefix: str = '',
        value_suffix: str = '',
        show_value_box: bool = False,
        value_box_color: _Optional[str] = None,
        value_box_width: _Optional[int] = None,
        value_box_height: _Optional[int] = None,
        value_box_y_shift: int = 15,  # extra arg. reminder: this parameter has a different argument name for horizontal and vertical sliders

        rod_len: int = 200,
        rod_thick: int = 3,
        btn_r: int = 5,
        color_rod_normal: str = '#4b4b4b',
        color_rod_locked: str = '#2d2d2d',
        color_btn_normal: str = '#555555',
        color_btn_locked: str = '#373737',
        color_btn_press: str = '#7d7d7d',
        color_btn_hover: str = '#696969',
        color_btn_bd_normal: str = '#f5f5f5',
        color_btn_bd_locked: str = '#373737',

        locked: bool = False,
        visible: bool = True,

        id: _Optional[str] = None,
        tags: _Optional[_Union[str, _List[str]]] = None,
    ) -> None:

        super().__init__(
            min, max, step, init, fn,
            x, y, anchor, tolerance,
            label, label_fg, label_font, show_label_box, label_box_color, label_box_width, label_box_height, label_y_shift,
            show_value, value_fg, value_font, value_prefix, value_suffix, show_value_box, value_box_color, value_box_width, value_box_height,
            rod_len, rod_thick, btn_r, color_rod_normal, color_rod_locked, color_btn_normal, color_btn_locked, color_btn_press, color_btn_hover, color_btn_bd_normal, color_btn_bd_locked,
            locked, visible,
            id, tags
        )

        self.value_box_y_shift = value_box_y_shift

        ## runtime (custom)

        self.width = self.rod_thick
        self.height = self.rod_len

        ## init        
        
        self._redraw()

    def _redraw(self):

        if self.locked:
            color_rod = self.color_rod_locked
            color_btn = self.color_btn_locked
            color_btn_bd = self.color_btn_bd_locked
        elif self.pressed:
            color_rod = self.color_rod_normal
            color_btn = self.color_btn_press
            color_btn_bd = self.color_btn_bd_normal
        elif self.hovered:
            color_rod = self.color_rod_normal
            color_btn = self.color_btn_hover
            color_btn_bd = self.color_btn_bd_normal
        else:
            color_rod = self.color_rod_normal
            color_btn = self.color_btn_normal
            color_btn_bd = self.color_btn_bd_normal

        _Slider._page.delete(f'Slider_{self.id}')

        if self.visible:

            X, Y = self.get_anchor_loc('center')
            w2 = self.width/2
            h2 = self.height/2

            _Slider._page.create_rectangle(
                X-w2, Y-h2,
                X+w2, Y+h2,
                fill=color_rod, width=0, tags=f'Slider_{self.id}'
            )
            _Slider._page.create_oval(
                X-self.btn_r, Y+h2-_norm(self.value, self.min, self.max)*self.height-self.btn_r,
                X+self.btn_r, Y+h2-_norm(self.value, self.min, self.max)*self.height+self.btn_r,
                fill=color_btn, width=1, outline=color_btn_bd, tags=f'Slider_{self.id}'
            )

            if self.show_label_box:
                _Slider._page.create_rectangle(
                    X-self.label_box_width/2, Y-h2+self.label_y_shift-self.label_box_height/2,
                    X+self.label_box_width/2, Y-h2+self.label_y_shift+self.label_box_height/2,
                    fill=self.label_box_color, width=0, tags=f'Slider_{self.id}'
                )
            if self.label is not None:
                _Slider._page.create_text(
                    X, Y-h2+self.label_y_shift,
                    text=self.label, font=self.label_font, fill=self.label_fg, tags=f'Slider_{self.id}'
                )

            if self.show_value_box:
                _Slider._page.create_rectangle(
                    X-self.value_box_width/2, Y+h2+self.value_box_y_shift-self.value_box_height/2,
                    X+self.value_box_width/2, Y+h2+self.value_box_y_shift+self.value_box_height/2,
                    fill=self.value_box_color, width=0, tags=f'Slider_{self.id}'
                )
            if self.show_value:
                _Slider._page.create_text(
                    X, Y+h2+self.value_box_y_shift,
                    text=self.value_prefix + str(self.value) + self.value_suffix,
                    font=self.value_font, fill=self.value_fg, tags=f'Slider_{self.id}'
                )

    def _hover(self):

        X, Y = self.get_anchor_loc('center')

        x = _Slider._page.winfo_pointerx()
        y = _Slider._page.winfo_pointery()

        bx = X
        by = Y + self.height/2 - _norm(self.value, self.min, self.max)*self.height
        br = self.btn_r

        inside = (bx-br <= x <= bx+br) and (by-br <= y <= by+br)

        if inside and (not self.locked) and self.visible and (not self.hovered):
            self.hovered = True
            self._redraw()
        
        elif self.hovered and (not inside):
            self.hovered = False
            self._redraw()

    def _press(self):

        X, Y = self.get_anchor_loc('center')

        x = _Slider._page.winfo_pointerx()
        y = _Slider._page.winfo_pointery()

        bx = X
        by = Y + self.height/2 - _norm(self.value, self.min, self.max)*self.height
        br = self.btn_r

        inside = (bx-br <= x <= bx+br) and (by-br <= y <= by+br)

        if inside and (not self.locked) and self.visible:
            self.pressed = True
            self._redraw()

    def _hold(self):

        if self.pressed:

            _, Y = self.get_anchor_loc('center')

            mousey = _Slider._page.winfo_pointery()
            value = self.max - ((mousey - (Y - self.height/2))/self.height)*(self.max - self.min)
    
            if self.prec == 0:
                value = int(value)
            else:
                value = round(value, self.prec)

            if abs(value - round(value/self.step)*self.step) < self.tolerance*self.step:

                if value < self.min:
                    value = self.min
                if value > self.max:
                    value = self.max
                
                if value == self.value:
                    return
                self.value = value
                
                self._redraw()
                if self.fn is not None:
                    self.fn()


import TurboTest as tt
from . import rcut


def Normal():

    tt.both_are_equal(rcut('foo12', '12'), 'foo')
    tt.both_are_equal(rcut('hi mom 123', ' 123'), 'hi mom')
    tt.both_are_equal(rcut('hi mom', 'cool'), 'hi mom')

    tt.both_are_equal(rcut('', 'ok'), '')
    tt.both_are_equal(rcut('abc', 'abc'), '')
    
    tt.both_are_equal(rcut('abc ', ' '), 'abc')
    
    tt.both_are_equal(rcut('abc', '123'), 'abc')
    tt.both_are_equal(rcut('oh my god', 'oh'), 'oh my god')

    ## Escaped chars
    tt.both_are_equal(rcut('a\nb', 'b'), 'a\n')
    tt.both_are_equal(rcut('a\nb', '\nb'), 'a')


def reject_empty_unwanted_arg():
    with tt.these_will_raise(ValueError) as its: rcut('hi', '')
    tt.both_are_equal(its.exception_msg, "rcut: `unwanted` shouldn't be an empty string.")


def Test_check():

    ## When False
    tt.both_are_equal(rcut('oh my god', '123'), 'oh my god')

    ## When True
    with tt.these_will_raise(AssertionError) as its: rcut('oh my god', '123', check=True)
    tt.both_are_equal(its.exception_msg, "rcut: the given `input_str` does not end with `unwanted`.")



import unittest

from mykit.ghactions.eLog import eL
from mykit.kit.stream_capture import StreamCapture


class Test__eL(unittest.TestCase):

    def setUp(self):
        eL._testing = True

    def test_set_level(self):
        
        ## Passes

        eL.set_level('quiet')
        eL.set_level('error')
        eL.set_level('Warning')
        eL.set_level('INFO')
        eL.set_level('debuG')

        ## Fails

        with self.assertRaises(ValueError) as ctx: eL.set_level('')
        self.assertEqual(str(ctx.exception), "Invalid level value: ''.")

        with self.assertRaises(ValueError) as ctx: eL.set_level('foo')
        self.assertEqual(str(ctx.exception), "Invalid level value: 'foo'.")

        with self.assertRaises(ValueError) as ctx: eL.set_level(0)
        self.assertEqual(str(ctx.exception), "Invalid level value: 0.")

    ## This test is failing, I think because it's been affected by
    ## the other tests? I'm not sure about the test-run order of
    ## unittest (plus inside GitHub Action VM).  ~TODO:Nicholas@20230810
    # def test_default(self):

    #     with StreamCapture() as captured:
    #         eL.group('group')
    #         eL.endgroup('endgroup')
    #         eL.debug('debug')
    #         eL.info('info')
    #         eL.warning('warning')
    #         eL.error('error')
    #     result = len( captured.value.split('\n') )  # Remember the '\n' at the end of the `captured.value` string
    #     expected = 7  # Count the number of lines, since the captured string contains 'clock' which can't be determined.
    #     self.assertEqual(result, expected)

    def test_at_debug_level(self):
        eL.set_level('DEBUG')
        with StreamCapture() as captured:
            eL.group('group')
            eL.endgroup('endgroup')
            eL.debug('debug')
            eL.info('info')
            eL.warning('warning')
            eL.error('error')
        self.assertEqual(len( captured.value.split('\n') ), 7)

    def test_at_info_level(self):
        eL.set_level('INFO')
        with StreamCapture() as captured:
            eL.group('group')
            eL.endgroup('endgroup')
            eL.debug('debug')
            eL.info('info')
            eL.warning('warning')
            eL.error('error')
        self.assertEqual(len( captured.value.split('\n') ), 6)

    def test_at_warning_level(self):
        eL.set_level('WARNING')
        with StreamCapture() as captured:
            eL.group('group')
            eL.endgroup('endgroup')
            eL.debug('debug')
            eL.info('info')
            eL.warning('warning')
            eL.error('error')
        self.assertEqual(len( captured.value.split('\n') ), 5)

    def test_at_error_level(self):
        eL.set_level('ERROR')
        with StreamCapture() as captured:
            eL.group('group')
            eL.endgroup('endgroup')
            eL.debug('debug')
            eL.info('info')
            eL.warning('warning')
            eL.error('error')
        self.assertEqual(len( captured.value.split('\n') ), 4)

    def test_at_quiet_level(self):
        eL.set_level('QUIET')
        with StreamCapture() as captured:
            eL.group('group')
            eL.endgroup('endgroup')
            eL.debug('debug')
            eL.info('info')
            eL.warning('warning')
            eL.error('error')
        self.assertEqual(len( captured.value.split('\n') ), 1)


if __name__ == '__main__':
    unittest.main()


import TurboTest as tt
from . import trim_long_str


def default_value():
    
    x  = 'a'*50
    x += 'b'*25
    x += 'c'*50
    
    result = trim_long_str(x)
    expected = 'a'*50 + '...\n\n   [25 more chars]\n\n...' + 'c'*50
    tt.both_are_equal(result, expected)


def return_the_original_string_if_its_length_is_not_exceeded():

    result = trim_long_str('hi')
    expected = 'hi'
    tt.both_are_equal(result, expected)


def Test_max__char():

    x  = 'a'*5
    x += 'b'*25
    x += 'c'*5    
    result = trim_long_str(x, max_char=10)
    expected = 'a'*5 + '...\n\n   [25 more chars]\n\n...' + 'c'*5
    tt.both_are_equal(result, expected)

    x  = 'a'*10
    x += 'b'*251
    x += 'c'*10
    result = trim_long_str(x, max_char=20)
    expected = 'a'*10 + '...\n\n   [251 more chars]\n\n...' + 'c'*10
    tt.both_are_equal(result, expected)

    x  = 'a'*150
    x += 'b'*1234
    x += 'c'*150
    result = trim_long_str(x, max_char=300)
    expected = 'a'*150 + '...\n\n   [1234 more chars]\n\n...' + 'c'*150
    tt.both_are_equal(result, expected)

    ## odd value max_char
    x  = 'a'*50
    x += 'b'*321
    x += 'c'*50
    result = trim_long_str(x, max_char=101)
    expected = 'a'*50 + '...\n\n   [321 more chars]\n\n...' + 'c'*50
    tt.both_are_equal(result, expected)


def Test_nTab():

    x  = 'a'*50
    x += 'b'*25
    x += 'c'*50
    
    result = trim_long_str(x, nTab=0)
    expected = 'a'*50 + '...\n\n[25 more chars]\n\n...' + 'c'*50
    tt.both_are_equal(result, expected)
    
    result = trim_long_str(x, nTab=2)
    expected = 'a'*50 + '...\n\n  [25 more chars]\n\n...' + 'c'*50
    tt.both_are_equal(result, expected)
    
    result = trim_long_str(x, nTab=5)
    expected = 'a'*50 + '...\n\n     [25 more chars]\n\n...' + 'c'*50
    tt.both_are_equal(result, expected)


def Test_with_escaped_chars():
    x = '12345678\nabc123456\n890'
    result = trim_long_str(repr(x), max_char=20)
    expected = "'12345678\\" + '...\n\n   [6 more chars]\n\n...' + "3456\\n890'"
    tt.both_are_equal(result, expected)


def Test_using_repr_and_return_original():
    x = 'hi 123'
    result = trim_long_str(repr(x))
    expected = repr(x)
    tt.both_are_equal(result, expected)



import enum as _enum
import subprocess as _sp

from mykit.kit.color import (
    Colored as _Colored,
    Hex as _Hex,
)
from mykit.kit.time import TimeFmt as _TimeFmt


class _Level(_enum.IntEnum):

    QUIET = 0

    _ADDONS = 50  # Used for the ones that will always be printed (echoed) at any level except QUIET

    ERROR = 100
    WARNING = 200
    INFO = 300
    DEBUG = 400


class eL:
    """
    eL (`echo` Log): A simple logger using `echo` function,
    intended for use within GitHub Action virtual machines.
    Inspired by `mykit.kit.pLog.pL`.
    """

    _testing = False  # Testing purposes

    level = _Level.DEBUG  # This is not a private variable (users should use it to get the current log level)

    @classmethod
    def set_level(cls, level:str, /) -> None:
        """
        `level`:
        - `'quiet'`
        - `'error'`
        - `'warning'`
        - `'info'`
        - `'debug'`
        """
        try:
            cls.level = getattr(_Level, str(level).upper())
        except AttributeError:
            raise ValueError(f'Invalid level value: {repr(level)}.')


    @classmethod
    def _echo(cls, text, level:_Level):
        if level <= cls.level:
            if cls._testing: print(text)
            else: _sp.run(['echo', text])


    @classmethod
    def group(cls, name:str, /) -> None:
        cls._echo(f'::group::{name}', _Level._ADDONS)

    @classmethod
    def endgroup(cls, name:str='', /) -> None:
        """
        ## Params
        - `name`: doesn't do anything, just for readability.
        
        ## Demo
        >>> eL.group('group-a')
        >>> eL.endgroup()
        >>> eL.group('group-b')
        >>> eL.endgroup('group-b')
        """
        cls._echo('::endgroup::', _Level._ADDONS)


    @classmethod
    def _logger(cls, level:_Level, color, msg):
        text = f'[{_TimeFmt.hour()}] {_Colored(level.name, color)}: {msg}'
        cls._echo(text, level)

    @classmethod
    def debug(cls, msg:str, /) -> None:
        cls._logger(_Level.DEBUG, _Hex.WHEAT, msg)

    @classmethod
    def info(cls, msg:str, /) -> None:
        cls._logger(_Level.INFO, _Hex.BLUE_GRAY, msg)

    @classmethod
    def warning(cls, msg:str, /) -> None:
        cls._logger(_Level.WARNING, _Hex.DARK_ORANGE, msg)

    @classmethod
    def error(cls, msg:str, /) -> None:
        cls._logger(_Level.ERROR, _Hex.SCARLET, msg)


import sys as _sys
from typing import (
    NoReturn as _NoReturn,
)

from mykit.kit.keycrate import KeyCrate as _KeyCrate
from mykit.kit.path import open_file as _open_file


class Monoconfig:
    """
    A simple plugin to handle a centralized, single set of settings for a program.

    Ideally, you will have this folder inside your program source:
    ```txt
    settings/
    └─ __init__.py
    └─ settings.txt

    # settings/__init__.py
    import os
    from mykit.kit.scpp.monoconfig import Monoconfig
    SETTINGS_FILE_PTH = os.path.join(os.path.dirname(__file__), 'settings.txt')
    cfg = Monoconfig(SETTINGS_FILE_PTH)
    ```
    """

    def __init__(self, file_path:str) -> None:
        """
        ## Params
        - `file_path`: Absolute path to the settings/configurations file

        ## Docs
        - There's minimal validation, such as not checking whether `file_path` exists
            or not (please handle that manually). TODO: Add validations for critical ones.
        """
        self.file_path = file_path

    def load_these_settings(self, *keys):
        """
        ## Usage
        >>> cfg = Monoconfig('settings.txt')
        >>> CFG = cfg.load_these_settings('foo', 'bar')
        >>> CFG.foo  # Accessing key `foo`
        >>> CFG.bar  # Accessing key `bar`
        >>> cfg.open()  # Open settings file then exit program
        """
        try:
            cfg = _KeyCrate(self.file_path, True, True, need_keys=keys)
        except (ValueError, FileNotFoundError, AttributeError, SyntaxError, AssertionError) as err:
            print(f"There are invalid configurations in the settings. Please open the software settings to fix this error: {err}")
            _sys.exit(1)
        return cfg

    def open(self) -> _NoReturn:
        """Open the settings file then exit."""
        _open_file(self.file_path)
        _sys.exit(0)





def time_quick(secs: float, /) -> str:
    """
    Convert seconds into minutes-seconds concise-look style

    ## Params
    - `secs`: in seconds (>= 0)

    ## Examples
    - `quick(61.25)` -> `'1m1.2s'`
    """
    m, s = divmod(abs(secs), 60)
    return f'{int(m)}m{s:.1f}s'



import json as _json
import os as _os
import subprocess as _sp
import sys as _sys
from typing import (
    Any as _Any
)


class SafeJSON:
    """Secure JSON read/write operations, ensuring data integrity during writing and rewriting."""

    @staticmethod
    def write(__pth: str, __obj: _Any, /, do_log: bool = True) -> None:

        ## normalize the path and perform some validations
        pth_norm = _os.path.normpath(__pth)
        if not pth_norm.lower().endswith('.json'):
            raise AssertionError(f'Not a JSON file: {repr(__pth)}.')
        if not _os.path.isdir(_os.path.dirname(pth_norm)):
            raise NotADirectoryError(f'The directory does not exist: {repr(__pth)}.')
        if _os.path.exists(pth_norm):
            raise FileExistsError(f'File already exists: {repr(__pth)}.')

        with open(pth_norm, 'w') as fp:
            _json.dump(__obj, fp)
        
        if do_log:
            print(f'INFO: Json written: {repr(__pth)}.')

    @staticmethod
    def read(__pth: str, /) -> _Any:

        ## normalize the path and perform some validations
        pth_norm = _os.path.normpath(__pth)
        if not pth_norm.lower().endswith('.json'):
            raise AssertionError(f'Not a JSON file: {repr(__pth)}.')
        if not _os.path.isfile(pth_norm):
            raise FileNotFoundError(f'Not a file: {repr(__pth)}.')

        with open(pth_norm, 'r') as fp:
            out = _json.load(fp)
        return out

    @staticmethod
    def rewrite(__pth: str, __obj: _Any, /, do_log: bool = True) -> None:

        ## normalize the path and perform some validations
        pth_norm = _os.path.normpath(__pth)
        if not pth_norm.lower().endswith('.json'):
            raise AssertionError(f'Not a JSON file: {repr(__pth)}.')
        if not _os.path.isfile(pth_norm):
            raise FileNotFoundError(f'Not a file: {repr(__pth)}.')

        tmp_file = pth_norm + '.tmp'
        bak_file = pth_norm + '.bak'
        if _os.path.exists(tmp_file):
            raise FileExistsError(f'Temporary file exists: {repr(tmp_file)}.')
        if _os.path.exists(bak_file):
            raise FileExistsError(f'Backup file exists: {repr(bak_file)}.')

        ## writing the new as temp
        with open(tmp_file, 'w') as fp:
            _json.dump(__obj, fp)

        _os.rename(pth_norm, bak_file)  # backup the previous
        _os.rename(tmp_file, pth_norm)  # rename temp to new
        _os.remove(bak_file)  # delete the previous

        if do_log:
            print(f'INFO: Json rewritten: {repr(__pth)}.')
    
    @staticmethod
    def recover(__pth: str, /) -> None:

        ## normalize and ensure that it's a JSON file
        pth_norm = _os.path.normpath(__pth)
        if not pth_norm.endswith('.json'):
            raise ValueError(f'Not a JSON file: {repr(__pth)}.')

        tmp_file = pth_norm + '.tmp'
        bak_file = pth_norm + '.bak'

        ## case I
        if _os.path.exists(pth_norm) and _os.path.exists(tmp_file) and (not _os.path.exists(bak_file)):
            _os.remove(tmp_file)
            return

        ## case II
        if (not _os.path.exists(pth_norm)) and _os.path.exists(tmp_file) and _os.path.exists(bak_file):
            _os.rename(tmp_file, pth_norm)
            _os.remove(bak_file)
            return

        ## case III
        if _os.path.exists(pth_norm) and (not _os.path.exists(tmp_file)) and _os.path.exists(bak_file):
            _os.remove(bak_file)
            return

        ## case IV (Weirdly happening, but currently disabled because I don't know if this is the case in other environments.)
        # if (not _os.path.exists(pth_norm)) and (not _os.path.exists(tmp_file)) and _os.path.exists(bak_file):
        #     _os.rename(bak_file, pth_norm)
        #     return


def open_file(file_pth: str, /) -> None:
    """
    Opens a file using the default system application.

    ---

    ## Exceptions
    - `NotImplementedError`: if the OS is unrecognizable
    """

    system = _sys.platform

    ## Windows
    if system == 'win32':
        _os.startfile(file_pth)

    ## macOS
    elif system == 'darwin':
        _sp.call(['open', file_pth])

    ## Linux
    elif system.startswith('linux'):
        _sp.call(['xdg-open', file_pth])

    ## Windows/Cygwin
    elif system.startswith('cygwin'):
        _sp.call(['cygstart', file_pth])

    else:
        raise NotImplementedError(f'Unsupported platform: {system}')


# import TurboTest as tt

# from mykit.kit.fileops.nice_json import NiceJSON

# import os
# import tempfile


# def _prepare():
#     pass

# def _clear():
#     pass


# def read_existing_file():
#     _prepare()
#     _clear()


# def read_nonexisting_file():
#     pass


# def bar():
    
#     tt.both_are_equal(myfunc(1, 2), 3)



# """

# [21:12:33] PASS: mykit/kit/path: read existing file  (120 ms)
# [21:12:33] FAIL: mykit/kit/path: read nonexisting file  (12 ms)
# [21:12:33] FAIL: mykit/kit/foo/bar: Test core 123  (3.2 secs)
# [21:12:33] FAIL: mykit/kit/foo/bar: Test core 123  (3 mins)

# """


import TurboTest as tt
from . import ctxt

from mykit.kit.color import Hex, hex_to_rgb


RED = '#ff0000'
GREEN = '#00ff00'
BLUE = '#0000ff'
YELLOW = '#ffff00'


# def Default_color():
#     r, g, b = hex_to_rgb(Hex.LAVENDER_GRAY)
#     result = repr(ctxt('hi mom'))
#     expected = f"'\\x1b[38;2;{r};{g};{b}mhi mom\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Single_color():
#     result = repr(ctxt('hi mom', RED))
#     expected = "'\\x1b[38;2;255;0;0mhi mom\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Single_color_and_a_plain_text():
#     result = repr(ctxt('hi mom', RED) + '123')
#     expected = "'\\x1b[38;2;255;0;0mhi mom\\x1b[0m123'"
#     tt.both_are_equal(result, expected)


# def A_plain_text_then_single_color():
#     result = repr('love you ' + ctxt('mom', RED))
#     expected = "'love you \\x1b[38;2;255;0;0mmom\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Colored_then_plain_then_colored():
#     result = repr(ctxt('love', RED) + ' love you ' + ctxt('mom', BLUE))
#     expected = "'\\x1b[38;2;255;0;0mlove\\x1b[0m love you \\x1b[38;2;0;0;255mmom\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Plain_then_colored_then_plain():
#     result = repr('hi ' + ctxt('love', GREEN) + ' wow')
#     expected = "'hi \\x1b[38;2;0;255;0mlove\\x1b[0m wow'"
#     tt.both_are_equal(result, expected)


# def Simple_color_in_color():
#     t1 = ctxt('hi', RED)
#     t2 = ctxt(t1 + 'mom' + t1, GREEN)
#     result = repr(t2)
#     expected = "'\\x1b[38;2;255;0;0mhi\\x1b[38;2;0;255;0mmom\\x1b[38;2;255;0;0mhi\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Complex_color_in_color():
#     t1 = ctxt('hi', RED)
#     t2 = ctxt(t1 + 'mom' + t1, GREEN)
#     t3 = ctxt(t2 + 'okay', BLUE)
#     result = repr(t3)
#     expected = "'\\x1b[38;2;255;0;0mhi\\x1b[38;2;0;255;0mmom\\x1b[38;2;255;0;0mhi\\x1b[38;2;0;0;255mokay\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Complex_color_in_color_II():
#     t1 = ctxt('hi', RED)
#     t2 = ctxt(t1 + 'mom' + t1, GREEN)
#     t3 = ctxt(t2 + 'okay', BLUE) + 'nocolor' + t1
#     t4 = ctxt(t3, '#ffff00')
#     result = repr(t4)
#     expected = "'\\x1b[38;2;255;0;0mhi\\x1b[38;2;0;255;0mmom\\x1b[38;2;255;0;0mhi\\x1b[38;2;0;0;255mokay\\x1b[38;2;255;255;0mnocolor\\x1b[38;2;255;0;0mhi\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Simple_color_with_background():
#     result = repr(ctxt('hi', RED, GREEN))
#     expected = "'\\x1b[38;2;255;0;0;48;2;0;255;0mhi\\x1b[0m'"
#     tt.both_are_equal(result, expected)


# def Simple_color_with_background_then_plain():
#     result = repr(ctxt('hi', RED, GREEN) + 'mom')
#     expected = "'\\x1b[38;2;255;0;0;48;2;0;255;0mhi\\x1b[0mmom'"
#     tt.both_are_equal(result, expected)


# def Complex_color_in_color_with_background():
#     t1 = ctxt('hi', RED, BLUE)
#     t2 = ctxt(t1 + 'mom' + t1, GREEN)
#     t3 = ctxt(t2 + 'okay', BLUE) + 'nocolor' + t1
#     t4 = ctxt(t3 + 'wow', '#ffff00', GREEN) + 'really-no-color'
#     result = repr(t4)
#     expected = "'\\x1b[38;2;255;0;0;48;2;0;0;255mhi\\x1b[38;2;0;255;0mmom\\x1b[38;2;255;0;0;48;2;0;0;255mhi\\x1b[38;2;0;0;255mokay\\x1b[38;2;255;255;0;48;2;0;255;0mnocolor\\x1b[38;2;255;0;0;48;2;0;0;255mhi\\x1b[38;2;255;255;0;48;2;0;255;0mwow\\x1b[0mreally-no-color'"
#     tt.both_are_equal(result, expected)


# def The_colored_cant_get_recolored():
#     pass



## to be honest, the above tests are so painful to read

FG_RED_NO_BG = '\x1b[38;2;255;0;0mA\x1b[0m'  # A
FG_GREEN_NO_BG = '\x1b[38;2;0;255;0mB\x1b[0m'  # B
FG_BLUE_NO_BG = '\x1b[38;2;0;0;255mC\x1b[0m'  # C
FG_RED_BG_YELLOW = '\x1b[38;2;255;0;0;48;2;255;255;0mD\x1b[0m'  # D
FG_YELLOW_NO_BG = '\x1b[38;2;255;255;0mE\x1b[0m'  # E
FG_BLUE_BG_GREEN = '\x1b[38;2;0;0;255;48;2;0;255;0mF\x1b[0m'  # F
if __name__ == '__main__':  # Debugging purposes
    import os;os.system('color')
    print('x ' + FG_RED_NO_BG + ' y')
    print('x ' + FG_GREEN_NO_BG + ' y')
    print('x ' + FG_BLUE_NO_BG + ' y')
    print('x ' + FG_RED_BG_YELLOW + ' y')
    print('x ' + FG_YELLOW_NO_BG + ' y')
    print('x ' + FG_BLUE_BG_GREEN + ' y')


def Default_color():
    r, g, b = hex_to_rgb(Hex.LAVENDER_GRAY)
    result = ctxt('hi mom')
    expected = f'\x1b[38;2;{r};{g};{b}mhi mom\x1b[0m'
    tt.both_are_equal(result, expected)


def Single_color():
    result = ctxt('A', RED)
    expected = FG_RED_NO_BG
    tt.both_are_equal(result, expected)


def Single_color_and_a_plain_text():
    result = ctxt('A', RED) + 'foo'
    expected = FG_RED_NO_BG + 'foo'
    tt.both_are_equal(result, expected)


def A_plain_text_then_single_color():
    result = 'foo' + ctxt('A', RED)
    expected = 'foo' + FG_RED_NO_BG
    tt.both_are_equal(result, expected)


def Colored_then_plain_then_colored():
    result = ctxt('A', RED) + 'foo' + ctxt('B', GREEN)
    expected = FG_RED_NO_BG + 'foo' + FG_GREEN_NO_BG
    tt.both_are_equal(result, expected)


def Plain_then_colored_then_plain():
    result = 'foo' + ctxt('A', RED) + 'bar'
    expected = 'foo' + FG_RED_NO_BG + 'bar'
    tt.both_are_equal(result, expected)


def Simple_color_in_color():
    t1 = ctxt('A', RED)
    t2 = ctxt(t1 + 'B' + t1, GREEN)
    result = t2
    expected = FG_RED_NO_BG + FG_GREEN_NO_BG + FG_RED_NO_BG
    tt.both_are_equal(result, expected)


def Complex_color_in_color():
    t1 = ctxt('A', RED)
    t2 = ctxt(t1 + 'B' + t1, GREEN)
    t3 = ctxt(t2 + 'C', BLUE)
    result = t3
    expected = FG_RED_NO_BG + FG_GREEN_NO_BG + FG_RED_NO_BG + FG_BLUE_NO_BG
    tt.both_are_equal(result, expected)


def Complex_color_in_color_II():
    t1 = ctxt('A', RED)
    t2 = ctxt(t1 + 'B' + t1, GREEN)
    t3 = ctxt(t2 + 'C', BLUE) + 'E' + t1
    t4 = ctxt(t3, YELLOW)
    result = t4
    expected = FG_RED_NO_BG + FG_GREEN_NO_BG + FG_RED_NO_BG + FG_BLUE_NO_BG + FG_YELLOW_NO_BG + FG_RED_NO_BG
    tt.both_are_equal(result, expected)


def Simple_color_with_background():
    result = ctxt('D', RED, YELLOW)
    expected = FG_RED_BG_YELLOW
    tt.both_are_equal(result, expected)


def Simple_color_with_background_then_plain():
    result = ctxt('D', RED, YELLOW) + 'foo'
    expected = FG_RED_BG_YELLOW + 'foo'
    tt.both_are_equal(result, expected)


def Complex_color_in_color_with_background():
    t1 = ctxt('D', RED, YELLOW)
    t2 = ctxt(t1 + 'B' + t1, GREEN)
    t3 = ctxt(t2 + 'C', BLUE) + 'F' + t1
    t4 = ctxt(t3 + 'F', BLUE, GREEN) + 'hi'
    result = t4
    expected = FG_RED_BG_YELLOW + FG_GREEN_NO_BG + FG_RED_BG_YELLOW + FG_BLUE_NO_BG + FG_BLUE_BG_GREEN + FG_RED_BG_YELLOW + FG_BLUE_BG_GREEN + 'hi'
    tt.both_are_equal(result, expected)


def The_colored_cant_get_recolored():
    t1 = ctxt('A', RED)
    t2 = ctxt(t1, GREEN)
    result = t2
    expected = FG_RED_NO_BG
    tt.both_are_equal(result, expected)

def The_colored_cant_get_recolored_II():
    t1 = ctxt('A', RED, YELLOW)
    t2 = ctxt(t1, GREEN)
    result = t2
    expected = FG_RED_BG_YELLOW
    tt.both_are_equal(result, expected)

def The_colored_cant_get_recolored_III():
    t1 = ctxt('A', RED)
    t2 = ctxt(t1, BLUE, GREEN)
    result = t2
    expected = FG_RED_NO_BG
    tt.both_are_equal(result, expected)



import unittest

from mykit.kit.py_file_compressor import py_file_compressor

import os

# from mykit.kit.


TEST_DATA = os.path.join(os.path.dirname(__file__), 'test_data')


class Test__py_file_compressor(unittest.TestCase):

    def test_set_1(self):

        return

        things = os.listdir(TEST_DATA)
        nThing = len(things)
        nSet = nThing//2  # number of sets

        ## Checks
        if (nThing % 2) != 0: raise AssertionError

        for i in range(nSet):

            pth_original = os.path.join(TEST_DATA, f'set_{i}_original.py')
            pth_compress = os.path.join(TEST_DATA, f'set_{i}_compress.py')
            
            py_file_compressor(pth_before)

            text_result = ''
            text_expected = ''

            self.assertEqual(text_result, text_expected)


if __name__ == '__main__':
    unittest.main()





## Birthdate: Oct 4, 2023
# def trim_long_str(input_str:str, /, max_char:int=100, nTab:int=3, make_it_raw:bool=True) -> str:
def trim_long_str(input_str:str, /, max_char:int=100, nTab:int=3) -> str:
    r"""
    Trim the `input_str` if its length is more than `max_char`

    ## Params
    - `input_str`: the input string. It's recommended to use `repr` if there are any
                    escaped characters to prevent them from being improperly truncated.
    - `max_char`: Let's say the string length is 100, but `max_char` is 35.
                    Then the remaining 65 characters (those in the middle of
                    the `input_str`) will be replaced by a placeholder "[X more chars]".
                    Note, even numbers are recommended.
    - `nTab`: the number of spaces before the placeholder (see examples below to learn more)

    ## Demo
    ```py
    x = 'super long string'
    print(trim_long_str(x))
    ```
    The output will look like this:
    ```txt
    aabbccddee...

       [2500 more chars]

    ...xxyyzz
    ```

    ## Demo II
    >>> print(trim_long_str(repr(SUPER_LONG_STR), nTab=os.get_terminal_size().columns//2))
    >>> aabbcc...
    >>>
    >>>                 [231 more chars]
    >>>
    >>> ...xx\nyyzz

    ## Docs
    - The reason why not using `os.get_terminal_size().columns//2` as the
        default `nTab` is because in a VM, it will raise an error since there is no terminal dimension.
    - There are minimal validations for function arguments. Please input appropriate values.
    - Also, using `nTab=os.get_terminal_size().columns//2` is a bit unreadable (makes it harder to find).  ~nicholas
    """

    str_len = len(input_str)

    ## Check
    if str_len <= max_char:
        return input_str
    
    ## Tweak: decrease max_char by 1 if it's an even number
    max_char = (max_char-1) if ( (max_char%2) != 0 ) else max_char

    remaining = str_len - max_char

    tab_str = ' '*nTab

    nHalf = int(max_char/2)

    return input_str[:nHalf] + f'...\n\n{tab_str}[{remaining} more chars]\n\n...' + input_str[-nHalf:]



import datetime as _datetime
import keyword as _keyword
import os as _os
import random as _random
import time as _time
from typing import (
    Dict as _Dict,
    Hashable as _Hashable,
    Tuple as _Tuple
)


def randfloat(__low: float, __high: float, __prec: int = 2, /) -> float:
    """
    This function follows `random.uniform` behaviors.
    `__prec`: float precision (>= 0); if `__low` at 2 decimal point (`0.25`), `__prec` should at least `2`.
    `__prec` normally max at 15-18 (depends on system).
    """

    ## can't use below because each end only half probability-dense
    # return round(_random.uniform(__low, __high), __prec)

    k = pow(10, __prec)
    return _random.randint(round(__low*k), round(__high*k)) / k


def randrange(low: float, high: float, len: float, /, pad: float = 0.1, prec: int = 3) -> _Tuple[float, float]:
    """
    if `low = 0, high = 1` -> include both `0` and `1`.
    if `low = 0, high = 1, pad = 0.1` -> include both `0.1` and `0.9`.

    `pad`: should `0 <= pad < 0.5`
    """

    range_len = high - low
    the_pad = range_len * pad

    start = randfloat(low + the_pad, high - the_pad - len, prec)
    end = start + len

    return (start, end)


def minmax_normalization(x: float, min: float, max: float, /) -> float:
    """min-max feature scaling"""
    return (x - min) / (max - min)


def slice_list(__in: list, __n: int, /) -> list:
    """if `__n = 2` -> `[1, 2, 3, 4, 5]` -> `[[1, 2], [3, 4], [5]]`"""
    out = [
        __in[i : i + __n]
        for i in range(0, len(__in), __n)
    ]
    return out


def map_range(__value, /, from_min, from_max, to_min, to_max) -> float:
    """
    Maps a value from one range to another.

    ---

    ## Params
    - `__value`: The value to be mapped
    - `from_min`: The minimum value of the original range
    - `from_max`: The maximum value of the original range
    - `to_min`: The minimum value of the target range
    - `to_max`: The maximum value of the target range

    ## Returns
    - The mapped value in the target range.

    ## Demo
    >>> original_value = 5
    >>> mapped_value = map_range(original_value, 1, 9, 0, 1)
    >>> print(mapped_value)
    0.5
    """
    
    ## normalize the value from the original range
    normalized_value = (__value - from_min) / (from_max - from_min)

    ## scale the normalized value to the target range
    mapped_value = normalized_value * (to_max - to_min) + to_min

    return mapped_value


def is_valid_var_name(__in: str, /) -> bool:
    """
    Check if a string `__in` is valid for variable name.

    ---

    ## Demo
    - `is_valid_var_name('2x')` -> `False`
    - `is_valid_var_name('x2')` -> `True`
    - `is_valid_var_name('cold-ice')` -> `False`
    - `is_valid_var_name('cold_ice')` -> `True`
    """
    return (__in.isidentifier() and (not _keyword.iskeyword(__in)))


def printer(__msg: str, /) -> None:
    """
    For simple logging needs.
    
    ---

    ## Demo
    >>> printer('INFO: foo')     # [06:15:09] INFO: foo
    >>> printer('WARNING: bar')  # [06:15:09] WARNING: bar
    """
    T = _datetime.datetime.now().strftime('%H:%M:%S')
    print(f'[{T}] {__msg}')

def slowprint(__msg: str, /, delay: float = 0.15) -> None:
    """
    For simple logging needs.
    Basically the same as `mykit.kit.utils.printer` but with delay.

    ---

    ## Params
    - `delay`: in seconds

    ## Demo
    >>> for i in range(3):
    >>>     slowprint(f'INFO: {i}')
    >>> ## output:
    >>> ## [06:38:32] INFO: 0
    >>> ## [06:38:32] INFO: 1
    >>> ## [06:38:33] INFO: 2
    """
    _time.sleep(delay)
    printer(__msg)

def print_screen(__msg: str, /) -> None:
    """prints message at the bottom of the terminal screen, adapting to terminal's height."""

    ## the height of users' terminal
    h = _os.get_terminal_size()[1]

    ## message height
    n = len( __msg.split('\n') )
    
    print( '\n'*(h-n) + __msg )


def sort_dict_by_key(input_dict:dict, /, reverse:bool=False) -> dict:
    """Sort dictionary by keys (keys should be sortable).

    ---
    
    ## Demo
    >>> sort_dict_by_key({'b': 3, 'a': 1, 'c': 2}, 0)  # {'a': 1, 'b': 3, 'c': 2}
    >>> sort_dict_by_key({'b': 3, 'a': 1, 'c': 2}, 1)  # {'c': 2, 'b': 3, 'a': 1}
    >>> sort_dict_by_key({3: 0, 1: 0, 2: 0}, 0)  # {1: 0, 2: 0, 3: 0}
    >>> sort_dict_by_key({3: 0, 1: 0, 2: 0}, 1)  # {3: 0, 2: 0, 1: 0}
    """
    out = {
        k: input_dict[k]
        for k in sorted(input_dict, reverse=reverse)
    }
    return out

def sort_dict_by_val(input_dict:dict, /, reverse:bool=False) -> dict:
    """Sort dictionary by values (values should be sortable).

    ---
    
    ## Demo
    >>> sort_dict_by_val({'x': 'b', 'y': 'a', 'z': 'c'}, 0)  # {'y': 'a', 'x': 'b', 'z': 'c'}
    >>> sort_dict_by_val({'x': 'b', 'y': 'a', 'z': 'c'}, 1)  # {'z': 'c', 'x': 'b', 'y': 'a'}
    >>> sort_dict_by_val({'a': 3, 'b': 1, 'c': 2}, 0)  # {'b': 1, 'c': 2, 'a': 3}
    >>> sort_dict_by_val({'a': 3, 'b': 1, 'c': 2}, 1)  # {'a': 3, 'c': 2, 'b': 1}
    """
    out = {
        k: input_dict[k]
        for k in sorted(input_dict, key=input_dict.get, reverse=reverse)
    }
    return out


def get_first_n_dict_items(input_dict:dict, /, first_n:int) -> dict:
    """
    Get the first `first_n` items from the dictionary.

    ---

    ## Params
    - `first_n`: should be at least 0. If it's negative, it'll be set to 0.
    """
    out = {
        k: input_dict[k]
        for k in list(input_dict)[:max(0, first_n)]
    }
    return out

def get_last_n_dict_items(input_dict:dict, /, last_n:int) -> dict:
    """
    Get the last `last_n` items from the dictionary.

    ---

    ## Params
    - `last_n`: should be at least 0. If it's negative, it'll be set to 0.
    """
    last_n = max(0, last_n)
    if last_n == 0: return {}
    return {
        k: input_dict[k]
        for k in list(input_dict)[-last_n:]
    }


def randhex(length:int=3, /) -> str:
    """
    Get a random hexadecimal string

    ---

    ## Demo
    >>> randhex()    # 13d
    >>> randhex()    # 3c4
    >>> randhex(7)   # f971445
    >>> randhex(21)  # 9aa4d78014ba3416dec3b
    """
    hex = '0123456789abcdef'
    return ''.join(_random.choices(hex, k=length))


def reverse_dict(input_dict: dict, /) -> dict:
    """
    Reverse the order of dict items

    ---

    ## Demo
    >>> reverse_dict({'a': 0, 'b': 0, 'c': 0})  # {'c': 0, 'b': 0, 'a': 0}
    >>> reverse_dict({5: '', 1: '', 9: ''})  # {9: '', 1: '', 5: ''}
    """
    keys = list(input_dict.keys())
    keys.reverse()
    return {
        k: input_dict[k]
        for k in keys
    }


def merge_dicts(
    dict1:_Dict[_Hashable, float],
    dict2:_Dict[_Hashable, float],
    /
) -> _Dict[_Hashable, float]:
    """
    Merge two dictionaries, returning the merged result,
    while keeping both input dictionaries `dict1` and `dict2` intact.

    ---

    ## Demo
    >>> dict1 = {'a': 1, 'b': 2}
    >>> dict2 = {'b': 3, 'c': 4}
    >>> merged_dict = merge_dicts(dict1, dict2)
    >>> # merged_dict is {'a': 1, 'b': 5, 'c': 4}

    ## Docs
    - See also `mykit.kit.utils.merging_dicts`
    """
    dict1 = dict1.copy()
    dict2 = dict2.copy()

    for key, value in dict2.items():
        if key in dict1:
            dict1[key] += value
        else:
            dict1[key] = value

    return dict1

def merging_dicts(
    dict1:_Dict[_Hashable, float],
    dict2:_Dict[_Hashable, float],
    /
) -> None:
    """
    Merge `dict2` into `dict1` (`dict2` unchanged).

    ---

    ## Demo
    >>> dict1 = {'a': 1, 'b': 2}
    >>> dict2 = {'b': 3, 'c': 4}
    >>> merging_dicts(dict1, dict2)
    >>> # dict1 will be {'a': 1, 'b': 5, 'c': 4}
    >>> # dict2 will be {'b': 3, 'c': 4}

    ## Docs
    - Also see `mykit.kit.utils.merge_dicts`
    """
    for key, value in dict2.items():
        if key in dict1:
            dict1[key] += value
        else:
            dict1[key] = value


def add_dict_val(
    the_dict:_Dict[_Hashable, float],
    key:_Hashable,
    add:float=1,
    init:float=1,
) -> None:
    """
    Add or initialize a value in the given dictionary

    ---

    ## Params
    - `the_dict`: A dictionary to which the value will be added or initialized.
    - `key`: The key for which the value will be modified or initialized.
    - `add`: The value to be added to the existing value associated with the key.
    - `init`: The initial value to set for the key if it doesn't exist.

    ## Demo
    >>> my_dict = {'apple': 5, 'banana': 3}
    >>> add_dict_val(my_dict, 'apple', 1, 1)
    >>> # After the function call, my_dict will be: {'apple': 6, 'banana': 3}
    >>> add_dict_val(my_dict, 'pear', 2, 0)
    >>> # After the function call, my_dict will be: {'apple': 6, 'banana': 3, 'pear': 0}
    """
    if key in the_dict:
        the_dict[key] += add
    else:
        the_dict[key] = init


import re as _re
import subprocess as _sp
from typing import (
    NoReturn as _NoReturn,
    Optional as _Optional,
    Union as _Union,
)


def run(commands: str, /, cwd:_Optional[str]=None) -> _Union[None, _NoReturn]:
    """
    A helper to use the `subprocess.run` function

    ## Examples
    - `run('echo 123')` is equal to `subprocess.run(['echo', '123'], cwd=None, check=True)`
    """
    cmds = _re.sub(r'[ ]+', ' ', commands.strip()).split(' ')
    _sp.run(cmds, cwd=cwd, check=True)


from mykit.kit.color import colored_len as _colored_len


def paragraph_width(text:str, /) -> str:
    """
    Return the length of the longest line in the given multi-line text `text` (also works if `text` is a single line).
    This also works for colored text produced by the mykit `Colored` function.
    """
    k = 0
    for line in text.split('\n'):
        l = _colored_len(line)
        if l > k:
            k = l
    return k



import io as _io
import sys as _sys
class StreamCapture:
 """Redirects and captures stream content"""
 def __init__(self,fd:int=1):
  r"""
Initialize the context manager.

---

## Params
- `fd`: file descriptors: `0` for stdin, `1` for stdout, and `2` is for stderr.

## Demo
>>> with StreamCapture() as captured:
>>>     print('foo\n\nbar')
>>>     print('baz')
>>> print(repr(captured.value))  # 'foo\n\nbar\nbaz\n'
  """
  self.fd = fd
  if fd in[0,2]:raise NotImplementedError
  if fd not in[0,1,2]:raise ValueError(f'Invalid `fd` value: {repr(fd)}.')
  self.holder=_io.StringIO()
  self.old=None
 def __enter__(self):
  if self.fd==0:self.old=_sys.stdin
  if self.fd==1:self.old=_sys.stdout
  if self.fd==2:self.old=_sys.stderr
  self.holder=self.holder
  if self.fd==0:_sys.stdin=self.holder
  if self.fd==1:_sys.stdout=self.holder
  if self.fd==2:_sys.stderr=self.holder
  return self
 def __exit__(self,exc_type,exc_value,traceback):
  if self.fd==0:_sys.stdin=self.old
  if self.fd==1:_sys.stdout=self.old
  if self.fd==2:_sys.stderr=self.old
 @property
 def value(self):
  """The captured output as a string"""
  return self.holder.getvalue()


import unittest

from mykit.kit.readable.paragraph_width import paragraph_width

from mykit.kit.color import Colored, Hex


class Test__paragraph_width(unittest.TestCase):

    def test_plain_text_single(self):
        
        text = ''
        result = paragraph_width(text)
        self.assertEqual(result, 0)
        
        text = 'hi'
        result = paragraph_width(text)
        self.assertEqual(result, 2)

    def test_plain_text_multi(self):
        
        text = 'hi\n123'
        result = paragraph_width(text)
        self.assertEqual(result, 3)
        
        text = (
            '123\n'
            '12345\n'
            '12\n'
            '1234567\n'
            '123'
        )
        result = paragraph_width(text)
        self.assertEqual(result, 7)

    def test_colored_text_single(self):
        
        text = Colored('hi')
        result = paragraph_width(text)
        self.assertEqual(result, 2)

    def test_colored_text_multi(self):
        
        text = (
            "12\n"
            f"{Colored('123', Hex.RED, Hex.BLACK)}\n"
            "1234\n"
            f"{Colored('12345')}\n"
            "1"
        )
        result = paragraph_width(text)
        self.assertEqual(result, 5)


if __name__ == '__main__':
    unittest.main()



import TurboTest as tt
from . import lcut


def Normal():
    tt.both_are_equal(lcut('121foo', '12'), '1foo')
    tt.both_are_equal(lcut('123 hi mom 123', '123 '), 'hi mom 123')
    tt.both_are_equal(lcut('hi mom', 'cool'), 'hi mom')

    tt.both_are_equal(lcut('', 'ok'), '')
    tt.both_are_equal(lcut('abc', 'abc'), '')
    
    tt.both_are_equal(lcut(' abc', ' '), 'abc')
    
    tt.both_are_equal(lcut('abc', '123'), 'abc')
    tt.both_are_equal(lcut('oh my god', 'god'), 'oh my god')

    ## Escaped chars
    tt.both_are_equal(lcut('a\nb', 'a'), '\nb')
    tt.both_are_equal(lcut('a\nb', 'a\n'), 'b')


def reject_empty_unwanted_arg():
    with tt.these_will_raise(ValueError) as its: lcut('hi', '')
    tt.both_are_equal(its.exception_msg, "lcut: `unwanted` shouldn't be an empty string.")


def Test_check():

    ## When False
    tt.both_are_equal(lcut('oh my god', '123'), 'oh my god')

    ## When True
    with tt.these_will_raise(AssertionError) as its: lcut('oh my god', '123', check=True)
    tt.both_are_equal(its.exception_msg, "lcut: the given `input_str` does not start with `unwanted`.")
